from __future__ import annotations

import math
import warnings
from itertools import count
from typing import TYPE_CHECKING, Callable, Sequence, Union

import numpy as np
import pandas as pd
from scipy.stats import chi2

from pykelihood.cached_property import cached_property

if TYPE_CHECKING:
    from pykelihood.distributions import Distribution

warnings.filterwarnings("ignore")


class ConditioningMethod(object):
    @staticmethod
    def no_conditioning(data: pd.Series, distribution: Distribution):
        return 0.0

    @staticmethod
    def excluding_last_obs_rule(data: pd.Series, distribution: Distribution):
        return distribution.logpdf(data.iloc[-1])

    @staticmethod
    def partial_conditioning_rule_stopped_obs(
        data: pd.Series, distribution: Distribution, threshold: Sequence = None
    ):
        return distribution.logsf(threshold[-1])

    @staticmethod
    def full_conditioning_rule_stopped_obs(
        data: pd.Series, distribution: Distribution, threshold: Sequence = None
    ):
        return distribution.logsf(threshold[-1]) + np.sum(
            distribution.logcdf(threshold[:-1])
        )


class Likelihood(object):
    def __init__(
        self,
        distribution: Distribution,
        data: pd.Series,
        conditioning_method: Callable = ConditioningMethod.no_conditioning,
        name: str = "Standard",
        inference_confidence: float = 0.99,
        fit_chi2: bool = False,
        single_profiling_param=None,
        compute_mle=True,
    ):
        """

        :param distribution: distribution on which the inference is based
        :param data: variable of interest
        :param conditioning_method: penalisation term for the likelihood
        :param name: name (optional) of the likelihood if it needs to be compared to other likelihood functions
        :param inference_confidence: wanted confidence for intervals
        :param fit_chi2: whether the results from the likelihood ratio method must be fitted to a chi2
        or a generic chi2 with degree of freedom 1 is used
        :param single_profiling_param: parameter that we want to fix to create the profiles based on likelihood
        """
        self.name = name
        self.distribution = distribution
        self.data = data
        self.conditioning_method = conditioning_method
        self.inference_confidence = inference_confidence
        self.fit_chi2 = fit_chi2
        self.single_profiling_param = single_profiling_param
        self.compute_mle = compute_mle

    @cached_property
    def standard_mle(self):
        estimate = self.distribution.fit(self.data)
        ll = estimate.log_likelihood(self.data)
        ll = ll if isinstance(ll, float) else ll[0]
        return (estimate, ll)

    @cached_property
    def mle(self):
        if self.compute_mle:
            x0 = self.distribution.optimisation_params
            estimate = self.distribution.fit_instance(
                self.data, penalty=self.conditioning_method, x0=x0
            )
        else:
            estimate = self.distribution
        ll_xi0 = estimate.log_likelihood(self.data, penalty=self.conditioning_method)
        ll_xi0 = ll_xi0 if isinstance(ll_xi0, float) else ll_xi0[0]
        return (estimate, ll_xi0)

    @cached_property
    def AIC(self):
        mle_aic = -2 * self.mle[1] + 2 * len(self.mle[0].optimisation_params)
        std_mle_aic = -2 * self.standard_mle[1] + 2 * len(
            self.standard_mle[0].optimisation_params
        )
        return {"AIC MLE": mle_aic, "AIC Standard MLE Fit": std_mle_aic}

    def Deviance(self):
        mle_deviance = -2 * self.mle[1]
        std_mle_deviance = -2 * self.standard_mle[1]
        return {
            "Deviance MLE": mle_deviance,
            "AIC Standard MLE Deviance": std_mle_deviance,
        }

    @cached_property
    def profiles(self):
        profiles = {}
        mle, ll_xi0 = self.mle
        if self.single_profiling_param is not None:
            params = [self.single_profiling_param]
        else:
            params = mle.optimisation_param_dict.keys()
        for name, k in mle.optimisation_param_dict.items():
            if name in params:
                r = k.real
                lb = r - 5 * (10 ** math.floor(math.log10(np.abs(r))))
                ub = r + 5 * (10 ** math.floor(math.log10(np.abs(r))))
                range = list(np.linspace(lb, ub, 50))
                profiles[name] = self.test_profile_likelihood(range, name)
        return profiles

    def test_profile_likelihood(self, range_for_param, param):
        mle, ll_xi0 = self.mle
        profile_ll = []
        params = []
        for x in range_for_param:
            try:
                pl = mle.fit_instance(
                    self.data,
                    penalty=self.conditioning_method,
                    fixed_params={param: x},
                )
                pl_value = pl.log_likelihood(
                    self.data, penalty=self.conditioning_method
                )
                pl_value = pl_value if isinstance(pl_value, float) else pl_value[0]
                if np.isfinite(pl_value):
                    profile_ll.append(pl_value)
                    params.append(list(pl.flattened_params))
            except:
                pass
        delta = [2 * (ll_xi0 - ll) for ll in profile_ll if np.isfinite(ll)]
        if self.fit_chi2:
            df, loc, scale = chi2.fit(delta)
            chi2_par = {"df": df, "loc": loc, "scale": scale}
        else:
            chi2_par = {"df": 1}
        lower_bound = ll_xi0 - chi2.ppf(self.inference_confidence, **chi2_par) / 2
        filtered_params = pd.DataFrame(
            [x + [ll] for x, ll in zip(params, profile_ll) if ll >= lower_bound]
        )
        cols = list(mle.flattened_param_dict.keys()) + ["likelihood"]
        filtered_params = filtered_params.rename(columns=dict(zip(count(), cols)))
        return filtered_params

    def confidence_interval(self, metric: Callable[[Distribution], float]):
        """

        :param metric: function depending on the distribution: it can be one of the parameter (ex: lambda x: x.shape() for a parameter called "shape"),
        or a metric relevant to the field of study (ex: the 100-years return level for extreme value analysis by setting lambda x: x.isf(1/100))...
        :return: bounds based on parameter profiles for this metric
        """
        estimates = []
        profiles = self.profiles
        if self.single_profiling_param is not None:
            params = [self.single_profiling_param]
        else:
            params = profiles.keys()
        for param in params:
            columns = list(self.mle[0].optimisation_param_dict.keys())
            result = profiles[param].apply(
                lambda row: metric(
                    self.distribution.with_params({k: row[k] for k in columns}.values())
                ),
                axis=1,
            )
            estimates.extend(list(result.values))
        if len(estimates):
            return [np.min(estimates), np.max(estimates)]
        else:
            return [None, None]


class DetrentedFluctuationAnalysis(object):
    def __init__(
        self,
        data: pd.DataFrame,
        scale_lim: Sequence[int] = None,
        scale_step: float = None,
    ):
        """

        :param data: pandas Dataframe, if it contains a column for the day and month, the profiles are normalized
        according to the mean for each calendar day averaged over years.
        :param scale_lim: limits for window sizes
        :param scale_step: steps for window sizes
        """
        if not ("month" in data.columns and "day" in data.columns):
            print("Will use the total average to normalize the data...")
            mean = data["data"].mean()
            std = data["data"].std()
            data = data.assign(mean=mean).assign(std=std)
        else:
            mean = (
                data.groupby(["month", "day"])
                .agg({"data": "mean"})["data"]
                .rename("mean")
                .reset_index()
            )
            std = (
                data.groupby(["month", "day"])
                .agg({"data": "std"})["data"]
                .rename("std")
                .reset_index()
            )
            data = data.merge(mean, on=["month", "day"], how="left").merge(
                std, on=["month", "day"], how="left"
            )
        phi = (data["data"] - data["mean"]) / data["std"]
        phi = (
            phi.dropna()
        )  # cases where there is only one value for a given day / irrelevant for DFA
        self.y = np.cumsum(np.array(phi))
        if scale_lim == None:
            lim_inf = 10 ** (math.floor(np.log10(len(data))) - 1)
            lim_sup = min(
                10 ** (math.ceil(np.log10(len(data)))), len(phi)
            )  # assuming all observations are equally splitted
            scale_lim = [lim_inf, lim_sup]
        if scale_step == None:
            scale_step = 10 ** (math.floor(np.log10(len(data)))) / 2
        self.scale_lim = scale_lim
        self.scale_step = scale_step

    @staticmethod
    def calc_rms(x: np.array, scale: int, polynomial_order: int):
        """
        windowed Root Mean Square (RMS) with polynomial detrending.
        Args:
        -----
          *x* : numpy.array
            one dimensional data vector
          *scale* : int
            length of the window in which RMS will be calculaed
        Returns:
        --------
          *rms* : numpy.array
            RMS data in each window with length len(x)//scale
        """
        # making an array with data divided in windows
        shape = (x.shape[0] // scale, scale)
        X = np.lib.stride_tricks.as_strided(x, shape=shape)
        # vector of x-axis points to regression
        scale_ax = np.arange(scale)
        rms = np.zeros(X.shape[0])
        for e, xcut in enumerate(X):
            coeff = np.polyfit(scale_ax, xcut, deg=polynomial_order)
            xfit = np.polyval(coeff, scale_ax)
            # detrending and computing RMS of each window
            rms[e] = np.mean((xcut - xfit) ** 2)
        return rms

    @staticmethod
    def trend_type(alpha: float):
        if round(alpha, 1) < 1:
            if round(alpha, 1) < 0.5:
                return "Anti-correlated"
            elif round(alpha, 1) == 0.5:
                return "Uncorrelated, white noise"
            elif round(alpha, 1) > 0.5:
                return "Correlated"
        elif round(alpha, 1) == 1:
            return "Noise, pink noise"
        elif round(alpha, 1) > 1:
            if round(alpha, 1) < 1.5:
                return "Non-stationary, unbounded"
            else:
                return "Brownian Noise"

    def __call__(
        self, polynomial_order: int, show=False, ax=None, supplement_title="", color="r"
    ):
        """
        Detrended Fluctuation Analysis - measures power law scaling coefficient
        of the given signal *x*.
        More details about the algorithm you can find e.g. here:
        Kropp, Jürgen, & Schellnhuber, Hans-Joachim. 2010. Case Studies. Chap. 8-11, pages 167–244 of : In extremis :
        disruptive events and trends in climate and hydrology. Springer Science & Business Media.
        """

        y = self.y
        scales = (
            np.arange(self.scale_lim[0], self.scale_lim[1], self.scale_step)
        ).astype(np.int)
        fluct = np.zeros(len(scales))
        # computing RMS for each window
        for e, sc in enumerate(scales):
            fluct[e] = np.sqrt(
                np.mean(self.calc_rms(y, sc, polynomial_order=polynomial_order))
            )
        # as this stage, F^2(s) should be something of the form s^h(2); taking the log should give a linear form of coefficient h(2)
        coeff = np.polyfit(np.log(scales), np.log(fluct), 1)
        # numpy polyfit returns the highest power first
        if show:
            import matplotlib

            matplotlib.rcParams["text.usetex"] = True
            ax = ax or matplotlib.pyplot.gca()
            default_title = "Detrended Fluctuation Analysis"
            title = (
                default_title
                if supplement_title == ""
                else f"{default_title} {supplement_title}"
            )
            fluctfit = np.exp(np.polyval(coeff, np.log(scales)))
            ax.loglog(scales, fluct, "o", color=color, alpha=0.6)
            ax.loglog(
                scales,
                fluctfit,
                color=color,
                alpha=0.6,
                label=r"DFA-{}, {}: $\alpha$={}".format(
                    polynomial_order, self.trend_type(coeff[0]), round(coeff[0], 2)
                ),
            )
            ax.set_title(title)
            ax.set_xlabel(r"$\log_{10}$(time window)")
            ax.set_ylabel(r"$\log_{10}$F(t)")
            ax.legend(loc="lower right", fontsize="small")
        return scales, fluct, coeff[0]


def pettitt(data: Union[np.array, pd.DataFrame, pd.Series]):
    """
    Pettitt's non-parametric test for change-point detection.
    Given an input signal, it reports the likely position of a single switch point along with
    the significance probability for location K, approximated for p <= 0.05.
    """
    T = len(data)
    X = data.reshape((len(data), 1))
    vector_of_ones = np.ones([1, len(X)])
    matrix_col_X = np.matmul(X, vector_of_ones)
    matrix_lines_X = matrix_col_X.T
    diff = matrix_lines_X - matrix_col_X
    diff_sign = np.sign(diff)
    U_initial = diff_sign[0, 1:].sum()
    sum_of_each_line = diff_sign[1:].sum(axis=1)
    cs = sum_of_each_line.cumsum()
    U = U_initial + cs
    U = list(U)
    U.insert(0, U_initial)
    loc = np.argmax(np.abs(U))
    K = np.max(np.abs(U))
    p = np.exp(-3 * K ** 2 / (T ** 3 + T ** 2))
    return (loc, p)



from typing import Dict


def to_tuple(x):
    if isinstance(x, str):
        return (x,)
    return tuple(x)


def flatten_dict(dict: Dict):
    res_dict = {}
    for k, v in dict.items():
        if not isinstance(v, Dict):
            res_dict[to_tuple(k)] = v
        else:
            new_dict = flatten_dict(v)
            for k_, v_ in new_dict.items():
                res_dict[to_tuple(k) + to_tuple(k_)] = v_
    return res_dict



import math
import warnings
from typing import List, Union

import matplotlib
import numpy as np
import pandas as pd

matplotlib.rcParams["text.usetex"] = True

from pykelihood import kernels
from pykelihood.distributions import Exponential, MixtureExponentialModel
from pykelihood.stats_utils import Likelihood

try:
    from hawkeslib import UnivariateExpHawkesProcess as UEHP
except ImportError:
    UEHP = None
from pykelihood.distributions import Distribution
from pykelihood.parameters import Parameter
from pykelihood.visualisation.utils import (
    get_quantiles_and_confidence_intervals,
    get_quantiles_and_confidence_intervals_uniform_scale,
)

warnings.filterwarnings("ignore")
import matplotlib.pyplot as plt


def qq_plot_gpd(
    data: pd.DataFrame,
    gpd_fit: Distribution,
    path_to_figure: str,
    threshold: Union[List, float, int, str] = "",
    ci_confidence=0.99,
    figure_name="qq_plot_gpd",
):
    id_obs = True
    if len(gpd_fit.flattened_params) != len(gpd_fit.params):
        id_obs = False
    theoretical, empirical, lower_bound, upper_bound = (
        get_quantiles_and_confidence_intervals(gpd_fit, data, ci_confidence)
        if id_obs
        else get_quantiles_and_confidence_intervals_uniform_scale(
            gpd_fit, data, ci_confidence
        )
    )
    n = len(data)
    text_title = ""
    if type(gpd_fit.loc()) is not Parameter:
        loc = {
            r"$\mu_{}$".format(a): round(gpd_fit.loc.param_dict[a], 2)
            for a in gpd_fit.loc.param_dict
        }
        for k in loc:
            text_title += f"{k} = {loc[k]}, "
    else:
        loc = round(gpd_fit.loc(), 2)
        text_title += r"$\mu$=" + str(loc) + ", "
    text_title += "\n"
    if type(gpd_fit.scale()) is not Parameter:
        scale = {
            r"$\sigma_{}$".format(a): round(gpd_fit.scale.param_dict[a], 2)
            for a in gpd_fit.scale.param_dict
        }
        for k in scale:
            text_title += f"{k} = {scale[k]}, "
    else:
        scale = round(gpd_fit.scale(), 2)
        text_title += r"$\sigma$=" + str(scale) + ", "
    text_title += "\n"
    if type(gpd_fit.shape()) is not Parameter:
        shape = {
            r"$\xi_{}$".format(a): round(gpd_fit.shape.param_dict[a], 2)
            for a in gpd_fit.shape.param_dict
        }
        for k in shape:
            text_title += f"{k} = {shape[k]}, "
    else:
        shape = round(gpd_fit.shape(), 2)
        text_title += r"$\xi$=" + str(shape)
    if text_title.endswith(", "):
        text_title = text_title[:-2]
    threshold_text = (
        str(tuple(threshold)) if hasattr(threshold, "__len__") else str(threshold)
    )
    plt.scatter(theoretical, empirical, s=5, marker="x", color="navy")
    plt.plot(theoretical, theoretical, label=f"$x=y$", color="navy")
    if id_obs:
        plt.fill_betweenx(
            y=empirical, x1=lower_bound, x2=upper_bound, alpha=0.2, color="navy"
        )
    else:
        plt.fill_between(theoretical, lower_bound, upper_bound, alpha=0.2, color="navy")
    plt.legend()
    plt.title(
        "QQ Plot of Exceedances over threshold "
        + threshold_text
        + " vs GPD distribution with parameters:\n"
        + text_title
    )
    plt.xlabel(f"Theoretical quantiles ({n} observations)")
    plt.ylabel("Empirical quantiles")
    plt.tight_layout()
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()
    to_concat = pd.DataFrame(
        [theoretical, lower_bound, upper_bound],
        columns=empirical.index,
        index=["theoretical", "lower_bound", "upper_bound"],
    ).T
    return pd.concat([empirical.rename("realized"), to_concat], axis=1)


def qq_plot_gev(
    data: pd.DataFrame,
    gev_fit: Distribution,
    path_to_figure: str,
    ci_confidence=0.99,
    figure_name="qq_plot_gev",
):
    id_obs = True
    if len(gev_fit.flattened_params) != len(gev_fit.params):
        id_obs = False
    theoretical, empirical, lower_bound, upper_bound = (
        get_quantiles_and_confidence_intervals(gev_fit, data, ci_confidence)
        if id_obs
        else get_quantiles_and_confidence_intervals_uniform_scale(
            gev_fit, data, ci_confidence
        )
    )
    n = len(data)
    text_title = ""
    if type(gev_fit.loc()) is not Parameter:
        loc = {
            r"$\mu_{}$".format(a): round(gev_fit.loc.param_dict[a], 2)
            for a in gev_fit.loc.param_dict
        }
        for k in loc:
            text_title += f"{k} = {loc[k]}, "
    else:
        loc = round(gev_fit.loc(), 2)
        text_title += r"$\mu$=" + str(loc) + ", "
    text_title += "\n"
    if type(gev_fit.scale()) is not Parameter:
        scale = {
            r"$\sigma_{}$".format(a): round(gev_fit.scale.param_dict[a], 2)
            for a in gev_fit.scale.param_dict
        }
        for k in scale:
            text_title += f"{k} = {scale[k]}, "
    else:
        scale = round(gev_fit.scale(), 2)
        text_title += r"$\sigma$=" + str(scale) + ", "
    text_title += "\n"
    if type(gev_fit.shape()) is not Parameter:
        shape = {
            r"$\xi_{}$".format(a): round(gev_fit.shape.param_dict[a], 2)
            for a in gev_fit.shape.param_dict
        }
        for k in shape:
            text_title += f"{k} = {shape[k]}, "
    else:
        shape = round(gev_fit.shape(), 2)
        text_title += r"$\xi$=" + str(shape)
    if text_title.endswith(", "):
        text_title = text_title[:-2]
    plt.scatter(theoretical, empirical, s=5, marker="x", color="navy")
    plt.plot(theoretical, theoretical, label=f"$x=y$", color="navy")
    if id_obs:
        plt.fill_betweenx(
            y=empirical, x1=lower_bound, x2=upper_bound, alpha=0.2, color="navy"
        )
    else:
        plt.fill_between(theoretical, lower_bound, upper_bound, alpha=0.2, color="navy")
    plt.title("QQ Plot of Maxima vs GEV distribution with parameters:\n" + text_title)
    plt.ylabel(f"Empirical ({n} observations)")
    plt.xlabel("GEV distribution")
    plt.tight_layout()
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()
    to_concat = pd.DataFrame(
        [theoretical, lower_bound, upper_bound],
        columns=empirical.index,
        index=["theoretical", "lower_bound", "upper_bound"],
    ).T
    return pd.concat([empirical.rename("realized"), to_concat], axis=1)


def consecutive_days_above_value_plot(
    data: pd.DataFrame, value: float, path_to_figure: str
):
    above_value = data[data["data"] >= value].assign(
        timedelta=lambda x: x["days_since_start"].diff()
    )
    consecutive = (
        above_value[above_value["timedelta"] == 1]
        .assign(timedelta=lambda x: x["days_since_start"].diff())
        .fillna(0.0)
        .assign(subgroup=lambda x: (x["timedelta"] != x["timedelta"].shift(1)).cumsum())
        .groupby(["year", "subgroup"])
        .agg({"data": "count"})
    )
    consecutive = consecutive + 1
    consecutive_mean_per_year = consecutive.groupby(level="year").agg("mean")
    quantile_inf = consecutive.groupby(level="year").agg(lambda x: np.quantile(x, 0.05))
    quantile_sup = consecutive.groupby(level="year").agg(lambda x: np.quantile(x, 0.95))
    plt.scatter(
        consecutive_mean_per_year.index,
        consecutive_mean_per_year,
        label="Mean",
        color="salmon",
        s=7,
    )
    plt.vlines(
        consecutive_mean_per_year.index,
        quantile_inf,
        quantile_sup,
        label=r"5\% Quantiles",
        alpha=0.6,
        color="salmon",
    )
    plt.legend()
    plt.title(f"Mean Number of Consecutive Days Above {value} per Year.")
    plt.savefig(f"{path_to_figure}/mean_nb_days_cons_above_{value}_year.png")
    plt.clf()


def consecutive_days_under_value_plot(
    data: pd.DataFrame, value: float, path_to_figure: str
):
    under_value = data[data["data"] <= value].assign(
        timedelta=lambda x: x["days_since_start"].diff()
    )
    consecutive = (
        under_value[under_value["timedelta"] == 1]
        .assign(timedelta=lambda x: x["days_since_start"].diff())
        .fillna(0.0)
        .assign(subgroup=lambda x: (x["timedelta"] != x["timedelta"].shift(1)).cumsum())
        .groupby(["year", "subgroup"])
        .agg({"data": "count"})
    )
    consecutive = consecutive + 1
    consecutive_mean_per_year = consecutive.groupby(level="year").agg("mean")
    quantile_inf = consecutive.groupby(level="year").agg(lambda x: np.quantile(x, 0.05))
    quantile_sup = consecutive.groupby(level="year").agg(lambda x: np.quantile(x, 0.95))
    plt.scatter(
        consecutive_mean_per_year.index,
        consecutive_mean_per_year,
        label="Mean",
        color="salmon",
        s=7,
    )
    plt.vlines(
        consecutive_mean_per_year.index,
        quantile_inf,
        quantile_sup,
        label=r"5\% Quantiles",
        alpha=0.6,
        color="salmon",
    )
    plt.legend()
    plt.title(f"Mean Number of Consecutive Days Under {value} per Year.")
    plt.savefig(f"{path_to_figure}/mean_nb_days_cons_under_{value}_year.png")
    plt.clf()


### INTER EXCEEDANCES DIAGNOSTIC PLOTS, CLUSTERING ###
def mean_inter_exceedance_time_per_year(data: pd.DataFrame, path_to_figure: str):
    """
    Plots the observed inter-exceedance times per year, useful to visualize an increase or decrease in the distance between two extreme events.
    :param data: Dataframe pandas with columns "data", "threshold" (possibly seasonal or periodical), "days_since_start" (of the full period) and "year".
    :param path_to_figure: Path to save the figure.
    :return: Plots diagnostic graphs.
    """
    data = data.dropna().reset_index()
    empty_years = []
    mean_year = []
    quantile_sup_year = []
    quantile_inf_year = []
    for year in data["year"].unique():
        try:
            iat_days = (
                data[(data["year"] == year) & (data["data"] >= data["threshold"])][
                    "days_since_start"
                ]
                .diff()
                .dropna()
            )
            quantile_sup_year.append(np.quantile(iat_days, q=0.99))
            quantile_inf_year.append(np.quantile(iat_days, q=0.01))
            mean_year.append(iat_days.mean())
        except:
            empty_years.append(year)
    not_y = [y for y in data["year"].unique() if y not in empty_years]
    plt.scatter(not_y, mean_year, s=7, color="salmon", label="Mean")
    plt.vlines(
        not_y,
        quantile_inf_year,
        quantile_sup_year,
        alpha=0.6,
        color="salmon",
        label=r"10\% Quantiles",
    )
    plt.title("Inter-exceedance time per year")
    plt.legend()
    plt.savefig(f"{path_to_figure}/inter_exceedances_per_year.png")
    plt.clf()


def qq_plot_exponential(
    data: Union[pd.Series, np.array],
    exp_fit: Distribution,
    path_to_figure: str,
    ci_confidence=0.99,
    figure_name="qq_plot_exponential",
):
    id_obs = True
    if len(exp_fit.flattened_params) != len(exp_fit.params):
        id_obs = False
    theoretical, empirical, lower_bound, upper_bound = (
        get_quantiles_and_confidence_intervals(exp_fit, data, ci_confidence)
        if id_obs
        else get_quantiles_and_confidence_intervals_uniform_scale(
            exp_fit, data, ci_confidence
        )
    )
    plt.scatter(theoretical, empirical, s=5, marker="x", color="navy")
    plt.plot(theoretical, theoretical, label=f"$x=y$", color="navy")
    plt.legend()
    if id_obs:
        plt.fill_betweenx(
            y=empirical, x1=lower_bound, x2=upper_bound, alpha=0.2, color="navy"
        )
    else:
        plt.fill_between(theoretical, lower_bound, upper_bound, alpha=0.2, color="navy")
    plt.title("QQ Plot of positive spacing vs Exponential distribution")
    plt.xlabel("Empirical")
    plt.ylabel(
        r"Exponential with parameter $\lambda$" + f"= {round(exp_fit.rate(), 2)}"
    )
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()
    to_concat = pd.DataFrame(
        [theoretical, lower_bound, upper_bound],
        columns=empirical.index,
        index=["theoretical", "lower_bound", "upper_bound"],
    ).T
    return pd.concat([empirical.rename("realized"), to_concat], axis=1)


def extremogram_plot(
    data: pd.DataFrame,
    h_range: Union[List, np.array],
    path_to_figure: str,
    compare_to_hawkes=False,
    figure_name="extremogram",
):
    """
    Plots the observed extremogram (ie proba of the observation in t+h to be an exceedance knowing that the observation in t was one.
    Compares the estimate using an homogeneous poisson process vs a Hawkes process.
    :param data: Dataframe pandas containing the columns "data" for the variable of interest, "threshold" for the (possibly seasonal) threshold(s) and "days_since_start" which is a non-normalized version
    of the time in days unit.
    :param h_range: Range for the h parameter to vary through.
    :param path_to_figure: Path to plot the extremogram.
    :return: Plots diagnostic graphs.
    """

    def extremogram_loop(h_range, indices_exceedances):
        counts = []
        for h in h_range:
            counts.append(
                len(np.intersect1d(indices_exceedances + h, indices_exceedances))
                / len(indices_exceedances)
            )
        return counts

    # empirical
    data_extremogram = data[["data", "days_since_start", "threshold"]]
    data_extremogram = (
        data_extremogram[data_extremogram["data"] >= data["threshold"]]
        .assign(iat=lambda x: x["days_since_start"].diff())
        .fillna(0.0)
    )
    indices_exceedances = [
        int(round(e, 0)) for e in list(data_extremogram["days_since_start"])
    ]
    extremogram_realized = extremogram_loop(h_range, indices_exceedances)
    extremogram_realized = pd.Series(extremogram_realized, index=h_range)

    # Simulated Poisson process
    exp_fit = Exponential.fit(data_extremogram["iat"], loc=0.0)
    exceedances_pp = []
    for i in range(1000):
        exceedances_pp.append(exp_fit.rvs(len(data_extremogram)).cumsum())
    count_pp = []
    for i in range(len(exceedances_pp)):
        # PP
        ex_pp = exceedances_pp[i].copy()
        indices_exceedances = [int(round(e, 0)) for e in ex_pp]
        local_count_pp = extremogram_loop(h_range, indices_exceedances)
        local_count_pp = pd.Series(local_count_pp, index=h_range)
        count_pp.append(local_count_pp)
    count_pp = pd.concat(count_pp, axis=1)
    mean_pp = count_pp.mean(axis=1)
    quantile_inf_pp = np.quantile(count_pp, q=0.01, axis=1)
    quantile_sup_pp = np.quantile(count_pp, q=0.99, axis=1)
    to_return = pd.concat(
        [
            extremogram_realized.rename("realized"),
            mean_pp.rename("mean_pp"),
            pd.DataFrame(
                [quantile_inf_pp, quantile_sup_pp],
                index=["pp_lb", "pp_ub"],
                columns=h_range,
            ).T,
        ],
        axis=1,
    )
    plt.bar(
        x=h_range,
        height=extremogram_realized,
        label="Empirical",
        width=0.8,
        color="slategrey",
        alpha=0.6,
    )
    plt.plot(h_range, mean_pp, label="Poisson Process Simulated", color="salmon")
    plt.fill_between(
        x=h_range, y1=quantile_inf_pp, y2=quantile_sup_pp, alpha=0.2, color="salmon"
    )

    # Simulated Hawkes process
    if compare_to_hawkes:
        if UEHP is not None:
            uv = UEHP()
            uv.fit(np.array(data_extremogram["days_since_start"]))
            mu, alpha, theta = uv.get_params()
        else:
            mu = len(data_extremogram) / len(data)
            alpha = 0
            theta = 0
        hawkes_fit = Exponential.fit(
            data_extremogram["iat"],
            x0=(0.0, mu, alpha, theta),
            rate=kernels.hawkes_with_exp_kernel(
                np.array(data_extremogram["days_since_start"])
            ),
        )
        exceedances_hp = []
        for i in range(1000):
            exceedances_hp.append(hawkes_fit.rvs(len(data_extremogram)).cumsum())
        count_hp = []
        for i in range(len(exceedances_hp)):
            # HP
            ex_hp = exceedances_hp[i]
            indices_exceedances = [int(round(e, 0)) for e in ex_hp]
            local_count_hp = extremogram_loop(h_range, indices_exceedances)
            local_count_hp = pd.Series(local_count_hp, index=h_range)
            count_hp.append(local_count_hp)
        count_hp = pd.concat(count_hp, axis=1)
        mean_hp = count_hp.mean(axis=1)
        quantile_inf_hp = np.quantile(count_hp, q=0.01, axis=1)
        quantile_sup_hp = np.quantile(count_hp, q=0.99, axis=1)
        to_return = pd.concat(
            [
                to_return,
                mean_hp.rename("mean_hp"),
                pd.DataFrame(
                    [quantile_inf_hp, quantile_sup_hp],
                    index=["hp_lb", "hp_ub"],
                    columns=h_range,
                ).T,
            ]
        )
        plt.plot(h_range, mean_hp, label="Hawkes Process Simulated", color="royalblue")
        plt.fill_between(
            x=h_range,
            y1=quantile_inf_hp,
            y2=quantile_sup_hp,
            alpha=0.3,
            color="royalblue",
        )

    plt.title(f"Extremogram")
    plt.xlabel(r"$h$")
    plt.ylabel(r"$\pi_h(u)$")
    plt.legend()
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()
    return to_return


def mean_cluster_size(
    data: pd.DataFrame,
    block_sizes: Union[List, np.array],
    path_to_figure: str,
    compare_to_hawkes=False,
    figure_name="mean_cluster_size",
):
    """
    Estimates the extremal intex by computing the mean cluster size considering blocks of sizes pre-defined in the parameter block_sizes, given that
    the block contains an exceedance (ie the mean of pi, the distribution of the extremal index).
    Compares the estimate using an homogeneous poisson process vs a Hawkes process.
    :param data: Dataframe pandas containing the columns "data" for the variable of interest, "threshold" for the defined (possibly seasonal) threshold(s) and "days_since_start" which is a non-normalized version
    of the time in days unit.
    :param block_sizes: Range for cluster sizes.
    :param path_to_figure: Path to save the figure.
    :return: Plots diagnostic graphs.
    """

    def loop_mean_cluster_size(
        indices_exceedances, length_observation_period, block_sizes
    ):
        local_count = []
        for bs in block_sizes:
            hp = np.histogram(
                indices_exceedances,
                bins=[bs * i for i in range(math.ceil(length_observation_period / bs))],
            )[0]
            hp = hp[hp > 0]
            local_count.append(np.mean(hp))
        local_count = pd.Series(local_count, index=block_sizes)
        return local_count

    data_extremal_index = data[["data", "days_since_start", "threshold"]]
    data_extremal_index = (
        data_extremal_index[
            data_extremal_index["data"] >= data_extremal_index["threshold"]
        ]
        .assign(iat=lambda x: x["days_since_start"].diff())
        .fillna(0.0)
    )
    # Empirical
    empirical_mean_cluster_size = []
    for bs in block_sizes:
        exceedances_clusters = np.histogram(
            data_extremal_index["days_since_start"],
            bins=[bs * i for i in range(math.ceil(len(data) / bs))],
        )[0]
        exceedances_clusters = exceedances_clusters[
            exceedances_clusters > 0
        ]  # given that the block contains an exceedance
        empirical_mean_cluster_size.append(np.mean(exceedances_clusters))
    empirical_mean_cluster_size = pd.Series(
        empirical_mean_cluster_size, index=block_sizes
    )

    # Simulated
    exp_fit = Exponential.fit(data_extremal_index["iat"], loc=0.0)
    exceedances_pp = []
    for i in range(1000):
        exceedances_pp.append(exp_fit.rvs(len(data_extremal_index)).cumsum())
    pp_cluster_sizes = []
    for i in range(len(exceedances_pp)):
        # PP
        ex_pp = exceedances_pp[i].copy()
        local_count_pp = loop_mean_cluster_size(
            ex_pp, data["days_since_start"].max(), block_sizes
        )
        pp_cluster_sizes.append(local_count_pp)
    pp_cluster_sizes = pd.concat(pp_cluster_sizes, axis=1)
    x = block_sizes
    mean_pp = pp_cluster_sizes.mean(axis=1)
    quantile_inf_pp = np.quantile(pp_cluster_sizes, q=0.01, axis=1)
    quantile_sup_pp = np.quantile(pp_cluster_sizes, q=0.99, axis=1)
    to_return = pd.concat(
        [
            empirical_mean_cluster_size.rename("realized"),
            mean_pp.rename("mean_pp"),
            pd.DataFrame(
                [quantile_inf_pp, quantile_sup_pp], index=["pp_lb", "pp_ub"], columns=x
            ).T,
        ],
        axis=1,
    )
    plt.bar(
        x=x,
        height=empirical_mean_cluster_size,
        label="Empirical",
        width=0.3,
        color="slategrey",
        alpha=0.6,
    )
    plt.plot(x, mean_pp, label="Poisson Process Simulated", color="salmon")
    plt.fill_between(
        x=x, y1=quantile_inf_pp, y2=quantile_sup_pp, alpha=0.2, color="salmon"
    )
    if compare_to_hawkes:
        if UEHP is not None:
            uv = UEHP()
            uv.fit(np.array(data_extremal_index["days_since_start"]))
            mu, alpha, theta = uv.get_params()
        else:
            mu = len(data_extremal_index) / len(data)
            alpha = 0
            theta = 0
        hawkes_fit = Exponential.fit(
            data_extremal_index["iat"],
            x0=(0.0, mu, alpha, theta),
            rate=kernels.hawkes_with_exp_kernel(
                np.array(data_extremal_index["days_since_start"])
            ),
        )
        exceedances_hp = []
        for i in range(1000):
            exceedances_hp.append(hawkes_fit.rvs(len(data_extremal_index)).cumsum())
        hp_cluster_sizes = []
        for i in range(len(exceedances_hp)):
            # HP
            ex_hp = exceedances_hp[i].copy()
            local_count_hp = loop_mean_cluster_size(
                ex_hp, data["days_since_start"].max(), block_sizes
            )
            hp_cluster_sizes.append(local_count_hp)
        hp_cluster_sizes = pd.concat(hp_cluster_sizes, axis=1)
        mean_hp = hp_cluster_sizes.mean(axis=1)
        quantile_inf_hp = np.quantile(hp_cluster_sizes, q=0.01, axis=1)
        quantile_sup_hp = np.quantile(hp_cluster_sizes, q=0.99, axis=1)
        to_return = pd.concat(
            [
                to_return,
                mean_hp.rename("mean_hp"),
                pd.DataFrame(
                    [quantile_inf_hp, quantile_sup_hp],
                    columns=x,
                    index=["hp_lb", "hp_ub"],
                ).T,
            ]
        )
        plt.plot(x, mean_hp, label="Hawkes Process Simulated", color="royalblue")
        plt.fill_between(
            x=x, y1=quantile_inf_hp, y2=quantile_sup_hp, alpha=0.3, color="royalblue"
        )
    plt.title(f"Mean Cluster Size")
    plt.xlabel(r"Block size $r$ (days)")
    plt.ylabel(r"Number of exceedances per block $\theta^{-1}_r(u)$")
    plt.legend(loc="upper left")
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()
    return to_return


def cum_number_exceedances(
    data: pd.DataFrame,
    length_total_period: Union[int, float],
    origin: pd.Timestamp,
    path_to_figure: str,
    compare_to_hawkes=False,
    figure_name="cum_number_of_exceedances",
):
    """
    Computes the observed and simulated cumulative number of exceedances and compares a homogeneous poisson process estimate with the one obtained
    using a Hawkes process for modeling inter-exceedance times.
    :param data: Dataframe pandas containing the columns "data", "threshold" and "time" (normalized between 0 and 1) for less numerical errors.
    :param length_total_period: Length of the total period for scaling (graphical purpose only, unlike in previous graphs
    comparing Hawkes and HomoPoisson where a discrete scale is preferred vs a normalized "continuous" one between 0 and 1 due
    to precision concerns using input block sizes and deltas), same unit as inter-exceedance times.
    :param origin: The start date for scaling purposes (graphical).
    :param path_to_figure: Path to save the figure.
    :return: Plots diagnostic graphs.
    """
    data_gpd = data.set_index("time")[["data", "threshold"]]
    data_gpd = (
        data_gpd[data_gpd["data"] >= data_gpd["threshold"]]
        .reset_index()
        .assign(iat=lambda x: x["time"].diff())
        .fillna(0.0)
    )
    # Empirical
    realized = data.set_index("time").assign(
        bool=lambda x: (x["data"] >= x["threshold"]).astype(int)
    )["bool"]
    realized = (
        realized[realized == True].cumsum().reset_index().set_index("bool")["time"]
    )

    def swap_axes(s):
        return pd.Series(s.index.values, index=s)

    realized_ = swap_axes(realized)
    x = pd.to_datetime(length_total_period * data_gpd["time"], unit="d", origin=origin)
    realized_.index = x

    # Simulated Poisson process
    exp_fit = Exponential.fit(data_gpd["iat"], loc=0.0)
    simulations = []
    for _ in range(1000):
        simulations.append(
            pd.DataFrame(np.cumsum(exp_fit.rvs(len(data_gpd))), columns=[_])
        )
    simulations = pd.concat(simulations, axis=1)
    mean_pp = simulations.mean(axis=1)
    quantile_inf_pp = np.quantile(simulations, q=0.01, axis=1)
    quantile_sup_pp = np.quantile(simulations, q=0.99, axis=1)
    mean_pp_ = swap_axes(mean_pp)
    x1_pp = pd.to_datetime(
        length_total_period * quantile_inf_pp, unit="d", origin=origin
    )
    x2_pp = pd.to_datetime(
        length_total_period * quantile_sup_pp, unit="d", origin=origin
    )
    mean_pp_.index = pd.to_datetime(
        length_total_period * mean_pp, unit="d", origin=origin
    )
    to_return = [realized_, mean_pp_]
    plt.plot(mean_pp_, label="Poisson Process Simulated", color="salmon")
    plt.fill_betweenx(y=mean_pp_, x1=x1_pp, x2=x2_pp, color="salmon", alpha=0.2)
    plt.plot(realized_, label="Empirical", color="slategrey")
    if compare_to_hawkes:
        if UEHP is not None:
            uv = UEHP()
            uv.fit(np.array(data_gpd["time"]))
            mu, alpha, theta = uv.get_params()
        else:
            mu = 1 / (len(data_gpd) / len(data))
            alpha = 0
            theta = 0
        hawkes_fit = Exponential.fit(
            data_gpd["iat"],
            x0=(0.0, mu, alpha, theta),
            rate=kernels.hawkes_with_exp_kernel(np.array(data_gpd["time"])),
        )
        simulations_hp = []
        for _ in range(1000):
            simulations_hp.append(
                pd.DataFrame(np.cumsum(hawkes_fit.rvs(len(data_gpd))), columns=[_])
            )
        simulations_hp = pd.concat(simulations_hp, axis=1)
        mean_hp = simulations_hp.mean(axis=1)
        quantile_inf_hp = np.quantile(simulations_hp, q=0.01, axis=1)
        quantile_sup_hp = np.quantile(simulations_hp, q=0.99, axis=1)
        mean_hp_ = swap_axes(mean_hp)
        x1 = pd.to_datetime(
            length_total_period * quantile_inf_hp, unit="d", origin=origin
        )
        x2 = pd.to_datetime(
            length_total_period * quantile_sup_hp, unit="d", origin=origin
        )
        mean_hp_.index = pd.to_datetime(
            length_total_period * mean_hp, unit="d", origin=origin
        )
        to_return.append(mean_hp_)
        plt.plot(mean_hp_, label="Hawkes Process Simulated", color="royalblue")
        plt.fill_betweenx(y=mean_hp_, x1=x1, x2=x2, color="royalblue", alpha=0.2)
    plt.title("Cumulative Exceedances over Threshold")
    plt.ylabel("Number of Exceedances")
    plt.xlabel("Time")
    plt.legend()
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()
    return to_return


def Kgaps_diagnostic_plots(
    range_K: Union[List, np.array],
    inter_exceedance_times: pd.Series,
    proba_of_exceedance: float,
    path_to_figure: str,
    ci_confidence=0.99,
):
    """

    :param range_K: Range in which the parameter K of the model varies; should be the same unit as the inter-exceedance times
    :param inter_exceedance_times: Observed inter-exceedance times
    :param proba_of_exceedance: Normalizing factor corresponding to 1-F(un) in the model, proba of exceeding the defined threshold
    :param path_to_figure: Path to save the plots
    :return: Plots the diagnostic graphs
    """
    Ks = range_K
    thetas = []
    CIs = []
    for K in Ks:
        iet_normalised_to_cluster_distance = (
            np.clip(inter_exceedance_times - K, 0.0, a_max=None) * proba_of_exceedance
        )
        mem = MixtureExponentialModel.fit(iet_normalised_to_cluster_distance)
        theta = mem.theta()
        ll = Likelihood(mem, iet_normalised_to_cluster_distance)
        CI = [ll.profiles["theta"]["theta"].min(), ll.profiles["theta"]["theta"].max()]
        thetas.append(theta)
        CIs.append(CI)
        freq, bins = np.histogram(iet_normalised_to_cluster_distance, bins="auto")
        freq = freq / len(iet_normalised_to_cluster_distance)
        x = (bins[:-1] + bins[1:]) / 2
        (
            theoretical,
            empirical,
            lower_bound,
            upper_bound,
        ) = get_quantiles_and_confidence_intervals(
            Exponential(rate=theta), iet_normalised_to_cluster_distance, ci_confidence
        )
        fig = plt.figure(figsize=(10, 5))
        gs = fig.add_gridspec(1, 2)
        ax = []
        for i in range(1):
            for j in range(2):
                axe = fig.add_subplot(gs[i, j])
                ax.append(axe)
        plt.subplots_adjust(wspace=0.2)
        ax0, ax1 = ax
        ax0.bar(x, freq, color="navy")
        ax0.set_title(
            f"Spacings between exceedances distribution for K={round(K, 1)} days"
        )
        ax0.set_xlabel("Interval")
        ax0.set_ylabel("Frequency")
        ax1.scatter(theoretical, empirical, s=5, color="navy")
        ax1.plot(theoretical, theoretical, label=f"$x=y$", color="navy")
        plt.fill_betweenx(
            y=empirical, x1=lower_bound, x2=upper_bound, alpha=0.5, color="navy"
        )
        ax1.legend()
        ax1.set_title("QQ Plot of positive spacing vs Exponential distribution")
        ax1.set_ylabel("Empirical")
        ax1.set_xlabel(r"Exponential with $\theta$" + f"= {round(theta, 2)}")
        fig.savefig(f"{path_to_figure}/positive_spacings_summary_{round(K, 1)}.png")
    serie = pd.Series([float(t) for t in thetas], index=Ks)
    CI = pd.DataFrame(CIs)
    plt.clf()
    plt.scatter(serie.index, serie, marker="x", color="navy")
    plt.vlines(serie.index, CI[0], CI[1], color="navy")
    plt.xlabel("Inter-cluster interval $K$ (days)")
    plt.ylabel(r"$\hat{\theta}$")
    plt.title("Extremal index estimate")
    plt.savefig(f"{path_to_figure}/K-gaps_extremal_index_estimate.png")
    plt.clf()



import warnings
from typing import Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

matplotlib.rcParams["text.usetex"] = True

from pykelihood.distributions import Distribution, Uniform
from pykelihood.stats_utils import Likelihood

warnings.filterwarnings("ignore")


def get_quantiles_and_confidence_intervals_uniform_scale(
    fit: Distribution,
    data: Union[pd.DataFrame, np.array, pd.Series],
    ci_confidence=0.99,
):
    ll = Likelihood(fit, data, inference_confidence=ci_confidence)
    min_max = []
    levels = np.linspace(0.01, 0.99, 100)
    for level in levels:
        metric = lambda x: pd.Series(x.cdf(data)).quantile(level)
        CI = ll.confidence_interval_for_specified_metric(metric)
        min_max.append(CI)
    min_max = pd.DataFrame(
        min_max, columns=["lower_bound", "upper_bound"], index=levels
    )
    empirical = pd.Series(ll.mle[0].cdf(data)).quantile(levels)
    theoretical = Uniform(0, 1).inverse_cdf(levels)
    return theoretical, empirical, min_max["lower_bound"], min_max["upper_bound"]


def get_quantiles_and_confidence_intervals(
    fit: Distribution,
    data: Union[pd.DataFrame, np.array, pd.Series],
    ci_confidence=0.99,
):
    ll = Likelihood(fit, data, inference_confidence=ci_confidence)
    min_max = []
    levels = np.linspace(0.01, 0.99, 100)
    empirical = pd.Series(data).quantile(levels)
    theoretical = fit.inverse_cdf(levels)
    for level in levels:
        metric = lambda x: x.inverse_cdf(level)
        CI = ll.confidence_interval_for_specified_metric(metric)
        min_max.append(CI)
    min_max = pd.DataFrame(
        min_max, columns=["lower_bound", "upper_bound"], index=levels
    )
    return theoretical, empirical, min_max["lower_bound"], min_max["upper_bound"]


def qq_plot_uniform_scale(
    fit: Distribution,
    data: Union[pd.DataFrame, np.array, pd.Series],
    path_to_figure: str,
    ci_confidence=0.99,
    figure_name="qq_plot",
):
    (
        theoretical,
        empirical,
        lower_bound,
        upper_bound,
    ) = get_quantiles_and_confidence_intervals_uniform_scale(fit, data, ci_confidence)
    n = len(data)
    plt.scatter(theoretical, empirical, s=5, color="navy")
    plt.plot(theoretical, theoretical, label=f"$x=y$", color="navy")
    plt.fill_between(theoretical, lower_bound, upper_bound, alpha=0.2, color="navy")
    plt.legend()
    plt.xlabel(f"Theoretical quantiles ({n} observations)")
    plt.ylabel("Empirical quantiles")
    plt.tight_layout()
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()


def qq_plot(
    fit: Distribution,
    data: Union[pd.DataFrame, np.array, pd.Series],
    path_to_figure: str,
    ci_confidence=0.99,
    figure_name="qq_plot",
):
    (
        theoretical,
        empirical,
        lower_bound,
        upper_bound,
    ) = get_quantiles_and_confidence_intervals(fit, data, ci_confidence)
    import matplotlib.pyplot as plt

    n = len(data)
    plt.scatter(theoretical, empirical, s=5, color="navy")
    plt.plot(theoretical, theoretical, label=f"$x=y$", color="navy")
    plt.fill_betweenx(
        y=empirical, x1=lower_bound, x2=upper_bound, alpha=0.2, color="navy"
    )
    plt.legend()
    plt.xlabel(f"Theoretical quantiles ({n} observations)")
    plt.ylabel("Empirical quantiles")
    plt.tight_layout()
    plt.savefig(f"{path_to_figure}/{figure_name}.png")
    plt.clf()



from abc import abstractmethod
from collections.abc import MutableSequence
from functools import partial, wraps
from typing import Any, Callable, Iterable, Union

import cachetools
import numpy as np
import pandas as pd
import scipy.special
from scipy.optimize import minimize
from scipy.stats import beta, expon, gamma, genextreme, genpareto, norm, pareto, uniform

from pykelihood import kernels
from pykelihood.parameters import ConstantParameter, Parametrized
from pykelihood.stats_utils import ConditioningMethod

EULER = -scipy.special.psi(1)


def hash_with_series(*args, **kwargs):
    to_hash = []
    for v in args:
        if isinstance(v, pd.Series):
            v = tuple(v.values)
        elif isinstance(v, MutableSequence) or isinstance(v, Iterable):
            v = tuple(v)
        to_hash.append(v)
    for k, v in kwargs.items():
        v = hash_with_series(v)
        to_hash.append((k, v))
    return hash(tuple(to_hash))


def ifnone(x, default):
    if x is None:
        return default
    return x


class Distribution(Parametrized):
    def __hash__(self):
        return (self.__class__.__name__,) + self.params

    @abstractmethod
    def rvs(self, size: int):
        return NotImplemented

    @abstractmethod
    def cdf(self, x: Union[np.array, float]):
        return NotImplemented

    @abstractmethod
    def isf(self, q: float):
        return NotImplemented

    @abstractmethod
    def pdf(self, x: Union[np.array, float]):
        return NotImplemented

    @classmethod
    def param_dict_to_vec(cls, x: dict):
        return tuple(x.get(p) for p in cls.params_names)

    def sf(self, x: Union[np.array, float]):
        return 1 - self.cdf(x)

    def logcdf(self, x: Union[np.array, float]):
        return np.log(self.cdf(x))

    def logsf(self, x: Union[np.array, float]):
        return np.log(self.sf(x))

    def logpdf(self, x: Union[pd.Series, np.array, float], *args, **kwds):
        return np.log(self.pdf(x, *args, **kwds))

    def inverse_cdf(self, q: float):
        return self.isf(1 - q)

    def log_likelihood(
        self,
        data: Union[np.array, pd.Series],
        penalty: Callable = ConditioningMethod.no_conditioning,
        *args,
        **kwds,
    ):
        res = self.logpdf(data, *args, **kwds)
        return np.sum(res) - penalty(data, self)

    def opposite_log_likelihood(
        self,
        data: Union[np.array, pd.Series],
        penalty: Callable = ConditioningMethod.no_conditioning,
        *args,
        **kwds,
    ):
        return -self.log_likelihood(data, penalty, *args, **kwds)

    def _process_fit_params(self, **kwds):
        param_dict = self.param_dict.copy()
        for name in param_dict:
            for name_fixed_param, value_fixed_param in kwds.items():
                if name_fixed_param == name:
                    param_dict[name] = ConstantParameter(value_fixed_param)
                elif name_fixed_param.startswith(name):
                    subparam = name_fixed_param.replace(f"{name}_", "")
                    subdic = param_dict[name].param_dict.copy()
                    subdic[subparam] = ConstantParameter(value_fixed_param)
                    param_dict[name] = param_dict[name].with_params(subdic.values())
        return param_dict

    @classmethod
    def fit(
        cls,
        data,
        x0=None,
        penalty=ConditioningMethod.no_conditioning,
        **fixed_values,
    ):
        init_parms = {}
        for k in cls.params_names:
            if k in fixed_values:
                v = fixed_values[k]
                if isinstance(v, Parametrized):
                    init_parms[k] = v
                else:
                    init_parms[k] = ConstantParameter(v)
        init = cls(**init_parms)
        x0 = x0 if x0 is not None else init.optimisation_params
        if len(x0) != len(init.optimisation_params):
            raise ValueError(
                f"Expected {len(init.optimisation_params)} values in x0, got {len(x0)}"
            )

        def to_minimize(x):
            o = init.with_params(x)
            return o.opposite_log_likelihood(data, penalty=penalty)

        res = minimize(to_minimize, x0, method="Nelder-Mead")
        return init.with_params(res.x)

    def fit_instance(
        self,
        data,
        penalty: Callable = ConditioningMethod.no_conditioning,
        fixed_params=None,
        **kwds,
    ):
        param_dict = self._process_fit_params(**(fixed_params or {}))
        kwds.update(param_dict)
        return self.fit(data, penalty=penalty, **kwds)


class AvoidAbstractMixin(object):
    def __getattribute__(self, item):
        x = object.__getattribute__(self, item)
        if (
            hasattr(x, "__isabstractmethod__")
            and x.__isabstractmethod__
            and hasattr(self, "__getattr__")
        ):
            x = self.__getattr__(item)
        return x


class ScipyDistribution(Distribution, AvoidAbstractMixin):
    base_module: Any

    def _correct_trends(self, f):
        @wraps(f)
        def g(*args, **kwargs):
            res = f(*args, **kwargs)
            res = np.array(res)
            # if res is a matrix, then we only want the diagonal
            try:
                x, y = res.shape
            except ValueError:
                # only 1 dimension
                return res
            else:
                if x == 1 or y == 1:
                    return res.flatten()
                if x != y:
                    raise ValueError("Unexpected size of arguments")
                return np.diag(res)

        return g

    def __getattr__(self, item):
        if item not in ("pdf", "logpdf", "cdf", "logcdf", "ppf", "isf", "sf", "logsf"):
            return super(ScipyDistribution, self).__getattr__(item)
        f = getattr(self.base_module, item)
        g = partial(self._wrapper, f)
        g = self._correct_trends(g)
        g = cachetools.cached(
            self._cache, key=lambda x: hash((item, hash_with_series(x)))
        )(g)
        self.__dict__[item] = g
        return g


class Uniform(ScipyDistribution):
    params_names = ("loc", "scale")
    base_module = uniform

    def __init__(self, loc=0.0, scale=1.0):
        super(Uniform, self).__init__(loc, scale)
        self._cache = {}

    def _wrapper(self, f, x, loc=None, scale=None):
        return f(x, ifnone(loc, self.loc()), ifnone(scale, self.scale()))

    def rvs(self, size, loc=None, scale=None):
        return uniform.rvs(ifnone(loc, self.loc()), ifnone(scale, self.scale()), size)


class Exponential(ScipyDistribution):
    params_names = ("loc", "rate")
    base_module = expon

    def __init__(self, loc=0.0, rate=1.0):
        super(Exponential, self).__init__(loc, rate)
        self._cache = {}

    def _wrapper(self, f, x, loc=None, rate=None):
        if rate is not None:
            rate = 1 / rate
        return f(x, ifnone(loc, self.loc()), ifnone(rate, 1 / self.rate()))

    def rvs(self, size, loc=None, rate=None):
        return self.base_module.rvs(
            ifnone(loc, self.loc()), ifnone(rate, 1 / self.rate()), size
        )


class Gamma(ScipyDistribution):
    params_names = ("loc", "scale", "shape")
    base_module = gamma

    def __init__(self, loc=0.0, scale=1.0, shape=0.0):
        super(Gamma, self).__init__(loc, scale, shape)
        self._cache = {}

    def _wrapper(self, f, x, loc=None, scale=None, shape=None):
        return f(
            x,
            ifnone(shape, self.shape()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
        )

    def rvs(self, size, loc=None, scale=None, shape=None):
        return self.base_module.rvs(
            ifnone(shape, self.shape()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
            size,
        )


class Pareto(ScipyDistribution):
    params_names = ("loc", "scale", "alpha")
    base_module = pareto

    def __init__(self, loc=0.0, scale=1.0, alpha=1.0):
        super(Pareto, self).__init__(loc, scale, alpha)
        self._cache = {}

    def _wrapper(self, f, x, loc=None, scale=None, alpha=None):
        return f(
            x,
            ifnone(alpha, self.alpha()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
        )

    def rvs(self, size, loc=None, scale=None, alpha=None):
        return self.base_module.rvs(
            ifnone(alpha, self.alpha()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
            size,
        )


class Beta(ScipyDistribution):
    params_names = ("loc", "scale", "alpha", "beta")
    base_module = beta

    def __init__(self, loc=0.0, scale=1.0, alpha=2.0, beta=1.0):
        super(Beta, self).__init__(loc, scale, alpha, beta)
        self._cache = {}

    def _wrapper(self, f, x, loc=None, scale=None, alpha=None, beta=None):
        return f(
            x,
            ifnone(alpha, self.alpha()),
            ifnone(beta, self.beta()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
        )

    def rvs(self, size, loc=None, scale=None, alpha=None, beta=None):
        return self.base_module.rvs(
            ifnone(alpha, self.alpha()),
            ifnone(beta, self.beta()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
            size,
        )


class Normal(ScipyDistribution):
    params_names = ("loc", "scale")
    base_module = norm

    def __init__(self, loc=0.0, scale=1.0):
        super(Normal, self).__init__(loc, scale)
        self._cache = {}

    def _wrapper(self, f, x, loc=None, scale=None):
        return f(x, ifnone(loc, self.loc()), ifnone(scale, self.scale()))

    def rvs(self, size, loc=None, scale=None):
        return self.base_module.rvs(
            ifnone(loc, self.loc()), ifnone(scale, self.scale()), size
        )


class GEV(ScipyDistribution):
    params_names = ("loc", "scale", "shape")
    base_module = genextreme

    def __init__(self, loc=0.0, scale=1.0, shape=0.0):
        super(GEV, self).__init__(loc, scale, shape)
        self._cache = {}

    def lb_shape(self, data):
        x_min = data.min()
        x_max = data.max()
        if x_min * x_max < 0:
            return -np.inf
        elif x_min > 0:
            return self.scale / (x_max - self.loc())
        else:
            return self.scale / (x_min - self.loc())

    def ub_shape(self, data):
        x_min = data.min()
        x_max = data.max()
        if x_min * x_max < 0:
            return np.inf
        elif x_min > 0:
            return self.scale / (x_min - self.loc())
        else:
            return self.scale / (x_max - self.loc())

    def _wrapper(self, f, x, loc=None, scale=None, shape=None):
        if shape is not None:
            shape = -shape
        return f(
            x,
            ifnone(shape, -self.shape()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
        )

    def rvs(self, size, shape=None, loc=None, scale=None):
        if shape is not None:
            shape = -shape
        return self.base_module.rvs(
            ifnone(shape, -self.shape()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
            size,
        )


class MixtureExponentialModel(Distribution):
    params_names = ("theta",)

    def __init__(self, theta=0.99):
        super(MixtureExponentialModel, self).__init__(theta)
        self._cache = {}

    def rvs(self, size):
        theta = self.theta()
        uniforms = Uniform().rvs(size)
        realisations = []
        for uniform in uniforms:
            if uniform <= 1 - theta:
                realisations.append(0.0)
            else:
                realisations.append(Exponential(rate=theta).rvs(1))
        return realisations

    def pdf(self, x):
        theta = self.theta()
        return np.where(x > 0.0, theta * Exponential(rate=theta).pdf(x), 1 - theta)

    def cdf(self, x):
        theta = self.theta()
        return np.where(x > 0.0, theta * Exponential(rate=theta).cdf(x), 1 - theta)

    def isf(self, q):
        theta = self.theta()
        return np.where(q != 1 - theta, Exponential(rate=theta).isf(q / theta), 0.0)


class GPD(ScipyDistribution):
    params_names = ("loc", "scale", "shape")
    base_module = genpareto

    def __init__(self, loc=0.0, scale=1.0, shape=0.0):
        super(GPD, self).__init__(loc, scale, shape)
        self._cache = {}

    def _wrapper(self, f, x, loc=None, scale=None, shape=None):
        return f(
            x,
            ifnone(shape, self.shape()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
        )

    def rvs(self, size, loc=None, scale=None, shape=None):
        return self.base_module.rvs(
            ifnone(shape, self.shape()),
            ifnone(loc, self.loc()),
            ifnone(scale, self.scale()),
            size,
        )


class ExtendedGPD(Distribution):
    params_names = ("loc", "scale", "shape")

    def __init__(self, loc=0.0, scale=1.0, shape=0.0):
        super(ExtendedGPD, self).__init__(loc, scale, shape)
        if shape == 0.0:
            raise ValueError(
                "For this distribution, the shape parameter must not be zero."
            )

    def rvs(self, size: int):
        return self.inverse_cdf(Uniform().rvs(size))

    def cdf(self, x: Union[np.array, float]):
        return 1 - np.clip(
            1 + self.shape * ((x - self.loc) / self.scale), 0, a_max=None
        ) ** (-1 / self.shape)

    def inverse_cdf(self, q: float):
        return self.loc + self.scale * ((1 - q) ** (-self.shape) - 1) / self.shape

    def pdf(self, x: Union[np.array, float]):
        return (1 / self.scale) * np.clip(
            1 + self.shape * ((x - self.loc) / self.scale), 0, a_max=None
        ) ** (-1 / self.shape - 1)

    def isf(self, q: float):
        return self.loc + self.scale * (q ** (-self.shape) - 1) / self.shape


class PointProcess(Distribution):
    def __init__(
        self,
        threshold: int,
        intensity,
        jumps_size_distribution: Distribution = GPD(),
        remaining_points_distribution: Distribution = Uniform(
            ConstantParameter(0.0), ConstantParameter(1.0)
        ),
        npp=365.25,
    ):
        """

        :param threshold: defines what is a jump in term of exceeding of a certain threshold
        :param jumps_size_distribution: the jump size distribution
        :param intensity: the inter-arrival intensity lambda which can depend on time
        (non homogeneous Poisson Process) or history (Hawkes Process).
         If not provided, Time is automatically set to indices and renormalized between 0 and 1.
        :param remaining_points_distribution: optional, handles the points that are not jumps (appearing with frequency
        1-lambda on average)
        :param npp: number of observations per period, used to compute return level estimates over k years
        """
        self.iat = Exponential(loc=ConstantParameter(0.0), rate=intensity)
        super(PointProcess, self).__init__(
            *(
                self.iat.params
                + jumps_size_distribution.params
                + remaining_points_distribution.params
            )
        )
        self.jumps_size = jumps_size_distribution
        self.remaining_points = remaining_points_distribution
        self.threshold = threshold
        self.intensity = intensity
        self.frequency_exceedances = 1 / intensity()
        self.npp = npp

    @property
    def params_names(self):
        IAT_names = [f"iat_{p}" for p in self.iat.params_names]
        JS_names = [f"js_{p}" for p in self.jumps_size.params_names]
        RemainingPoints_names = [f"rp_{p}" for p in self.remaining_points.params_names]
        return tuple(IAT_names) + tuple(JS_names) + tuple(RemainingPoints_names)

    def with_params(self, new_params):
        new_params = iter(new_params)
        iat = self.iat.with_params(new_params)
        js = self.jumps_size.with_params(new_params)
        rp = self.remaining_points.with_params(new_params)
        return type(self)(self.threshold, iat.rate, js, rp, self.npp)

    def _process_fit_params(self, **kwds):
        IAT_kwds = {
            k.replace("iat_", ""): v for (k, v) in kwds.items() if k.startswith("iat_")
        }
        JS_kwds = {
            k.replace("js_", ""): v for (k, v) in kwds.items() if k.startswith("js_")
        }
        RP_kwds = {
            k.replace("rp_", ""): v for (k, v) in kwds.items() if k.startswith("rp_")
        }
        IAT_params = self.iat._process_fit_params(**IAT_kwds)
        JS_params = self.jumps_size._process_fit_params(**JS_kwds)
        RP_params = self.remaining_points._process_fit_params(**RP_kwds)
        if len(kwds) > len(IAT_kwds) + len(JS_kwds) + len(RP_kwds):
            raise ValueError(f"Unexpected parameters encountered in keywords: {kwds}")
        return IAT_params, JS_params, RP_params

    def fit_instance(
        self,
        data,
        penalty: Callable = ConditioningMethod.no_conditioning,
        fixed_params=None,
        **kwds,
    ):
        kwds.update(**(fixed_params or {}))
        return self.fit(data, penalty=penalty, **kwds)

    def fit(
        self,
        data,
        penalty=ConditioningMethod.no_conditioning,
        x0=None,
        opt_method="Nelder-Mead",
        plot=False,
        *args,
        **kwds,
    ):
        params_iat, params_js, params_rp = self._process_fit_params(**kwds)
        obj = type(self)(
            self.threshold,
            type(self.iat)(**params_iat).rate,
            type(self.jumps_size)(**params_js),
            type(self.remaining_points)(**params_rp),
        )
        # Three different fittings to avoid numerical instability
        inter_arrival_times = np.diff(
            np.asarray(data >= self.threshold).nonzero()[0] / len(data)
        )
        inter_arrival_times = np.insert(inter_arrival_times, 0, 0)
        distribution_iat = obj.iat.fit(inter_arrival_times, **obj.iat.param_dict)
        pot = data[data >= self.threshold]
        distribution_jump_size = obj.jumps_size.fit(pot, **obj.jumps_size.param_dict)
        remaining_points = data[(~data.isin(pot)) & data > 0]
        distribution_remaining_points = obj.remaining_points.fit(
            remaining_points, **obj.remaining_points.param_dict
        )
        res = list(
            tuple(distribution_iat.optimisation_params)
            + tuple(distribution_jump_size.optimisation_params)
            + tuple(distribution_remaining_points.optimisation_params)
        )
        fitted_distribution = obj.with_params(res)
        fitted_distribution.frequency_exceedances = (
            fitted_distribution.intensity() / len(data)
        )
        return fitted_distribution

    def show_diagnostic_plot(
        self, data, time_scale=("d", pd.to_datetime("1961-01-01"))
    ):
        """

        :param data: empirical observations used for the fit
        :param time_scale: tuple (a,b) with a being the frequency of observations per period (ie daily data over a year:
        a="d") and b the first period (for data starting in year 1961, b=pd.to_datetime("1961-01-01"))
        :return: plots
        """
        import matplotlib.pyplot as plt

        n_simul = 500
        simul, exceedances = self.rvs(
            n_simul, fixed_number_of_exceedances=len(data[data >= self.threshold])
        )
        fig = plt.figure(figsize=(10, 10))
        gs = fig.add_gridspec(2, 2)
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[1, 0])
        ax3 = fig.add_subplot(gs[0, 1])
        ax4 = fig.add_subplot(gs[1, 1])

        # Cumulative Number of Exceedances
        simu = pd.concat([pd.Series(d.index) for d in exceedances], axis=1)
        pot = data[data >= self.threshold]
        realized = (pot >= self.threshold).cumsum()
        simu.index = simu.index + 1
        mean = simu.mean(axis=1)
        std = simu.std(axis=1)

        def swap_axes(s):
            return pd.Series(s.index.values, index=s)

        mean_ = swap_axes(mean)
        x = pd.to_datetime(
            len(data) * realized.index, unit=time_scale[0], origin=time_scale[1]
        )
        x1 = pd.to_datetime(
            len(data) * mean_.index, unit=time_scale[0], origin=time_scale[1]
        )
        x2 = pd.to_datetime(
            len(data) * (mean + 2 * std), unit=time_scale[0], origin=time_scale[1]
        )
        x3 = pd.to_datetime(
            len(data) * (mean - 2 * std), unit=time_scale[0], origin=time_scale[1]
        )
        realized.index = x
        mean_.index = x1
        ax1.plot(realized, label="Realized", color="black")
        ax1.plot(mean_, label=f"Simulated ({n_simul} points)", color="r")
        ax1.fill_betweenx(y=mean_, x1=x2, x2=x3, color="r", alpha=0.2)
        ax1.set_title("Cumulative Exceedances over Threshold")
        ax1.set_ylabel("Number of Exceedances")
        ax1.set_xlabel("Time")
        ax1.legend()

        # Peaks Over Threshold
        realized_cdf = pot.value_counts(normalize=True).sort_index().cumsum()
        if len(self.jumps_size.params) == len(self.jumps_size.flattened_params):
            theoretical = self.jumps_size.cdf(realized_cdf.index)
            ax2.plot(realized_cdf, theoretical)
            ax2.plot(realized_cdf, realized_cdf)
            ax2.set_title("Quantile Plot Peaks Over Threshold")
            ax2.set_xlabel("Empirical")
            ax2.set_ylabel("Theoretical")
        else:
            simulated = pd.concat(
                [
                    e.transform(lambda x: round(x, 1), axis=0)
                    .value_counts(normalize=True)
                    .sort_index()
                    .cumsum()
                    for e in exceedances
                ],
                axis=1,
            )
            mean = (
                pd.concat([simulated.mean(axis=1), realized_cdf], axis=1)
                .sort_index()
                .ffill()
                .loc[realized_cdf.index][0]
            )
            ax2.plot(realized_cdf, mean)
            ax2.plot(realized_cdf, realized_cdf)
            ax2.set_title("Quantile Plot Peaks Over Threshold")
            ax2.set_xlabel("Empirical")
            ax2.set_ylabel(f"Simulated ({n_simul} points)")

        # Points Under Threshold
        ut = data[data < self.threshold]
        realized_cdf = ut.value_counts(normalize=True).sort_index().cumsum()
        theoretical = self.remaining_points.cdf(realized_cdf.index)
        ax3.plot(realized_cdf, theoretical)
        ax3.plot(realized_cdf, realized_cdf)
        ax3.set_title("Quantile Plot Points Under Threshold")
        ax3.set_xlabel("Empirical")
        ax3.set_ylabel("Theoretical")

        # Return-level graph
        nb_period = int(round(len(data) / self.npp, 0))
        simulated_points = pd.concat(simul, axis=1)
        simulated_points.index = nb_period * simulated_points.index
        data_rescaled = data.copy()
        data_rescaled.index = nb_period * data_rescaled.index
        if not isinstance(self.intensity(), float):
            freq_series = pd.Series(
                self.frequency_exceedances,
                index=self.intensity.f.args[0],
                name="intensity",
            )
            freq_series.index = nb_period * freq_series.index
        else:
            freq_series = pd.Series(
                [self.frequency_exceedances] * len(pot),
                index=nb_period * pot.index,
                name="intensity",
            )
        true_vs_realized = []
        rle = self.isf(1.0)
        for i, return_level_estimate in zip(freq_series.index, rle):
            true_return_level = data_rescaled.loc[i - 1 : i].max()
            simulated_mean = simulated_points.loc[i - 1 : i].max().mean()
            simulated_std = simulated_points.loc[i - 1 : i].std().mean()
            true_vs_realized.append(
                pd.DataFrame(
                    [
                        [
                            true_return_level,
                            return_level_estimate,
                            simulated_mean,
                            simulated_mean - 2 * simulated_std,
                            simulated_mean + 2 * simulated_std,
                        ]
                    ],
                    columns=[
                        "Realized",
                        "Estimated",
                        f"Simulated ({n_simul} Points)",
                        "SimulBInf",
                        "SimulBSup",
                    ],
                    index=[
                        time_scale[1] + pd.to_timedelta(f"{self.npp*i} {time_scale[0]}")
                    ],
                )
            )
        true_vs_realized = pd.concat(true_vs_realized)

        def moving_average(x):
            ret = np.cumsum(x)
            bool = (x > 0).cumsum()
            return ret / bool

        for col, color in zip(true_vs_realized.columns, ["black", "g", "r", "", ""]):
            true_vs_realized[col] = moving_average(true_vs_realized[col])
            if col not in ["SimulBInf", "SimulBSup"]:
                ax4.plot(true_vs_realized[col], label=col, color=color)
        inf_bound, sup_bound = (
            true_vs_realized["SimulBInf"],
            true_vs_realized["SimulBSup"],
        )
        ax4.fill_between(
            x=true_vs_realized.index, y1=inf_bound, y2=sup_bound, color="r", alpha=0.2
        )
        ax4.set_title("1 year Return level Moving Average")
        ax4.set_xlabel("Time")
        ax4.set_ylabel("Return level")
        ax4.legend()
        return fig

    def rvs(self, size, fixed_number_of_exceedances=None, show_diagnostic_plots=True):
        from pykelihood.samplers import HawkesByThinningModified, PoissonByThinning

        simul = []
        exceedances_raw = []
        for _ in range(size):
            if fixed_number_of_exceedances:
                nb_exceedances = fixed_number_of_exceedances
                expo = np.cumsum(self.iat.rvs(size=nb_exceedances))
            else:
                if (
                    self.intensity.original_f.func.__name__
                    == kernels.hawkes_with_exp_kernel.__name__
                ):
                    mu, alpha, theta = self.intensity.params
                    expo = HawkesByThinningModified(1.0, mu, alpha, theta)
                else:
                    expo = PoissonByThinning(
                        1.0, self.intensity, np.max(self.intensity())
                    )
                nb_exceedances = len(expo)
            exceedances = np.array(self.jumps_size.rvs(nb_exceedances))
            exceedances_raw.append(pd.Series(exceedances, index=expo, name=f"{_}"))
            total_number_of_points = int(
                np.mean(self.intensity()) / np.mean(self.frequency_exceedances)
            )
            non_exceedances = total_number_of_points - nb_exceedances
            under_threshold_points = self.remaining_points.rvs(size=non_exceedances)
            indices_utp = np.random.choice(
                list(set(np.linspace(0, 1, total_number_of_points)).difference(expo)),
                non_exceedances,
                replace=False,
            )
            points_over_threshold = pd.Series(exceedances, index=expo, name=f"{_}")
            points_under_threshold = pd.Series(
                under_threshold_points, index=indices_utp, name=f"{_}"
            )
            simul.append(
                pd.concat([points_under_threshold, points_over_threshold]).sort_index()
            )
        return simul, exceedances_raw

    def cdf(self, x):
        freq = self.frequency_exceedances
        if not isinstance(x, pd.Series) and not isinstance(freq, float):
            raise TypeError(
                "The information of the covariate value is missing, please input a pandas Series"
                "with index corresponding to this information."
            )
        if isinstance(x, pd.Series) and not isinstance(freq, float):
            freq_series = pd.Series(
                freq / len(x), index=self.intensity.f.args[0], name="intensity"
            )
            freq = (
                pd.concat([freq_series, x], axis=1)
                .sort_index()
                .bfill()
                .ffill()["intensity"]
            )
        res = np.where(
            x < self.threshold,
            self.remaining_points.cdf(x) * (1 - freq),
            freq * self.jumps_size.cdf(x),
        )
        return res

    def isf(self, q):
        freq = self.frequency_exceedances
        return_level_proba = q / self.npp
        res = np.where(
            return_level_proba > freq,
            self.remaining_points.isf(return_level_proba / (1 - freq)),
            self.jumps_size.isf(return_level_proba / freq),
        )
        return res

    def pdf(self, x):
        freq = self.frequency_exceedances
        if not isinstance(x, pd.Series) and not isinstance(freq, float):
            raise TypeError(
                "The information of the covariate value is missing, please input a pandas Series"
                "with index corresponding to this information."
            )
        if isinstance(x, pd.Series) and not isinstance(freq, float):
            freq_series = pd.Series(
                freq, index=self.intensity.f.args[0], name="intensity"
            )
            freq = (
                pd.concat([freq_series, x], axis=1)
                .sort_index()
                .bfill()
                .ffill()["intensity"]
            )
        res = np.where(
            x < self.threshold,
            (1 - freq) * self.remaining_points.pdf(x),
            self.jumps_size.pdf(x) * freq,
        )
        return res


class CompositionDistribution(Distribution):
    def __init__(self, d1: Distribution, d2: Distribution):
        """
        Distribution of cdf GoF
        :param d1: distribution aimed at replacing the uniform sampling in the formula used
        to generate variables with distribution d2 (G in F^{-1}(G^{-1}(U))
        :param d2: main objective distribution, especially aimed at determining the asymptotic behaviour.
        """
        super(CompositionDistribution, self).__init__(*(d1.params + d2.params))
        self.d1 = d1
        self.d2 = d2

    @property
    def params_names(self):
        d1_names = [f"d1_{p}" for p in self.d1.params_names]
        d2_names = [f"d2_{p}" for p in self.d2.params_names]
        return tuple(d1_names) + tuple(d2_names)

    def with_params(self, new_params):
        new_params = iter(new_params)
        d1 = self.d1.with_params(new_params)
        d2 = self.d2.with_params(new_params)
        return type(self)(d1, d2)

    def rvs(self, size):
        return self.d2.inverse_cdf(self.d1.rvs(size))

    def cdf(self, x):
        return self.d1.cdf(self.d2.cdf(x))

    def isf(self, q):
        return self.d2.isf(self.d1.isf(1 - q))

    def pdf(self, x, *args):
        return self.d2.pdf(x) * self.d1.pdf(self.d2.cdf(x))

    def _process_fit_params(self, **kwds):
        d1_kwds = {
            k.replace("d1_", ""): v for (k, v) in kwds.items() if k.startswith("d1_")
        }
        d2_kwds = {
            k.replace("d2_", ""): v for (k, v) in kwds.items() if k.startswith("d2_")
        }
        d1_params = self.d1._process_fit_params(**d1_kwds)
        d2_params = self.d2._process_fit_params(**d2_kwds)
        if len(kwds) > len(d1_kwds) + len(d2_kwds):
            raise ValueError(f"Unexpected parameters encountered in keywords: {kwds}")
        return d1_params, d2_params

    def fit_instance(
        self,
        data,
        penalty: Callable = ConditioningMethod.no_conditioning,
        fixed_params=None,
        **kwds,
    ):
        kwds.update(**(fixed_params or {}))
        return self.fit(data, penalty=penalty, **kwds)

    def fit(
        self,
        data: pd.Series,
        penalty=ConditioningMethod.no_conditioning,
        x0=None,
        opt_method="Nelder-Mead",
        *args,
        **kwds,
    ):
        params_d1, params_d2 = self._process_fit_params(**kwds)
        obj = type(self)(type(self.d1)(**params_d1), type(self.d2)(**params_d2))
        x0 = obj.optimisation_params if x0 is None else x0

        def to_minimize(var_params):
            return obj.with_params(var_params).opposite_log_likelihood(data, penalty)

        result = minimize(
            to_minimize,
            method=opt_method,
            options={"adaptive": True, "maxfev": 1000},
            x0=x0,
        )
        res = list(result.x)
        return obj.with_params(res)



from typing import Callable

import numpy as np

from pykelihood.distributions import Exponential, Uniform


def PoissonByThinning(T, lambd, M):
    """
    :param T: Maximum time
    :param lambd: Function or float describing the intensity of the Poisson Process
    :param M: Upper bound on [0, T] for lambd(.)
    :return: sample timestamps in ascending order
    """
    if not isinstance(lambd, Callable):
        functional_lambd = lambda x: lambd
    else:
        functional_lambd = lambd
    P = []
    t = 0
    while t < T:
        e = Exponential(loc=0.0, rate=M).rvs(1)
        t += e
        u = Uniform(loc=0.0, scale=M).rvs(1)
        if t < T and u <= functional_lambd(t):
            P.append(float(t))
    return P


def HawkesByThinning(T, lambd, tau=None):
    """
    From article https://arxiv.org/pdf/1507.02822.pdf
    :param T: Maximum time
    :param lambd: Intensity function, must be non-decreasing inbetween arrival times
    :param tau: true inter-arrival times, optional, if not provided the simulated ones are used
    :return: sample timestamps in ascending order
    """
    epsilon = 1e-10
    P = []
    tau = P if tau is None else tau
    t = 0
    while t < T:
        M = lambd(t + epsilon, tau)
        e = Exponential(rate=M).rvs(1)
        t += e
        u = Uniform(scale=M).rvs(1)
        if t < T and u <= lambd(t, tau):
            P.append(float(t))
    return P


def HawkesByThinningModified(T, mu, alpha, theta, phi=0):
    """
    Implements Ogata's modified thinning algorithm for sampling from a univariate Hawkes process
    with exponential decay.
    :param T: the maximum time
    :param mu: the exogenous intensity
    :param alpha: the infectivity factor alpha
    :param theta: intensity parameter of the delay density
    :param phi: (optionally) the starting phi, the running sum of exponential differences specifying the history until
    a certain time, thus making it possible to take conditional samples
    :return: 1-d numpy ndarray of samples
    """
    t = 0.0
    d = 0.0
    P = []
    while t < T:
        M = mu + alpha * theta * (1 + phi)
        e = Exponential(rate=M).rvs(1)
        t += e
        exp_decay = np.exp(-theta * (e + d))
        lambda_ = mu + alpha * theta * exp_decay * (1 + phi)
        u = Uniform(loc=0.0, scale=M).rvs(1)
        if t < T and u <= lambda_:
            P.append(float(t))
            phi = exp_decay * (1 + phi)
            d = 0
        else:
            d += e
    return P



import re
from itertools import count
from typing import Collection, Sequence, Union

import numpy as np
import pandas as pd

from pykelihood.parameters import ConstantParameter, Parameter, ParametrizedFunction


def parametrized_function(**param_defaults):
    def wrapper(f):
        def wrapped(*args, **param_values):
            final_params = {
                p_name: Parameter(v) for p_name, v in param_defaults.items()
            }
            final_params.update(
                {p_name: ConstantParameter(v) for p_name, v in param_values.items()}
            )
            return ParametrizedFunction(f, *args, **final_params)

        return wrapped

    return wrapper


# Usual kernels with one covariate
@parametrized_function(a=0.0, b=0.0)
def linear(X, a, b):
    return a + b * X


@parametrized_function(a=0.0, b=0.0, c=0.0)
def polynomial(X, a, b, c):
    return a + b * X + c * X ** 2


@parametrized_function(a=0.0, b=0.0)
def expo(X, a, b):
    inner = b * X
    inner = a + inner
    return np.exp(inner)


@parametrized_function(mu=0.0, sigma=1.0, scaling=0.0)
def gaussian(X, mu, sigma, scaling):
    mult = scaling * 1 / (sigma * np.sqrt(2 * np.pi))
    expo = np.exp(-((X - mu) ** 2) / sigma ** 2)
    return mult * expo


@parametrized_function(mu=0.0, alpha=0.0, theta=1.0)
def hawkes_with_exp_kernel(X, mu, alpha, theta):
    return mu + alpha * theta * np.array(
        [np.sum(np.exp(-theta * (X[i] - X[:i]))) for i in range(len(X))]
    )


def hawkes2(t, tau, mu, alpha, theta):
    return mu + alpha * theta * np.sum(
        (np.exp(-theta * (t - tau[i]))) for i in range(len(tau)) if tau[i] < t
    )


# Sophisticated kernels with multiple covariates
def linear_regression(
    x: Union[int, pd.DataFrame, np.ndarray] = 2, add_intercept=False, **constraints
) -> ParametrizedFunction:
    """Computes a trend as a linear sum of the columns in the data.

    :param x: the number of dimensions or the data the kernel will be computed on. There will be one parameter for each column.
    :param add_intercept: if True, an intercept is added to the result
    :param constraints: fixed values for the parameters of the regression. The constraints are given as `beta_i=value`,
                        where i is the index of the column starting with 1.
                        If data is given as a dataframe with the second column named 'cname', the following constraints are equivalent:
                        'beta_2=2', 'beta_cname=2', 'cname=2'.
                        'beta_0' constrains the value of the intercept if add_intercept is True.
    """
    args = ()
    if isinstance(x, int):
        assert x > 0, "Unexpected number of parameters for linear regression"
        ndim = x
    else:
        args = (x,)
        if len(x.shape) > 1:
            ndim = x.shape[1]
        else:
            raise ValueError(
                "Consider using kernels.linear for a 1-dimensional data array"
            )
    fixed = {}
    for p_name, p_value in constraints.items():
        if p_name.startswith("beta_"):
            p_name = p_name[len("beta_") :]
        if isinstance(x, pd.DataFrame) and p_name in x.columns:
            index = tuple(x.columns).index(p_name) + 1
        else:
            index = int(p_name)
        fixed[index] = ConstantParameter(p_value)

    if 0 in fixed and not add_intercept:
        raise ValueError(
            "A fixed value is given for the intercept, but `add_intercept` is not True."
        )

    param_indices = range(0 if add_intercept else 1, ndim + 1)
    params = {
        f"beta_{i}": Parameter() if i not in fixed else fixed[i] for i in param_indices
    }

    def _compute(data, **params_from_wrapper):
        intercept = params_from_wrapper.pop("beta_0", 0)
        sorted_params = [params_from_wrapper[k] for k in params if k != "beta_0"]
        return intercept + (sorted_params * data).sum(axis=1)

    return ParametrizedFunction(
        _compute, *args, **params, fname=linear_regression.__qualname__
    )


def exponential_linear_regression(
    x: Union[int, pd.DataFrame, np.ndarray] = 2, add_intercept=False, **constraints
) -> ParametrizedFunction:
    """Computes a trend as the exponential of a linear sum of the columns in the data.

    :param x: the number of dimensions or the data the kernel will be computed on. There will be one parameter for each column.
    :param add_intercept: if True, an intercept is added to the result
    :param constraints: fixed values for the parameters of the regression. The constraints are given as `beta_i=value`,
                        where i is the index of the column starting with 1.
                        If data is given as a dataframe with the second column named 'cname', the following constraints are equivalent:
                        'beta_2=2', 'beta_cname=2', 'cname=2'.
                        'beta_0' constrains the value of the intercept if add_intercept is True.
    """
    args = ()
    if isinstance(x, int):
        assert x > 0, "Unexpected number of parameters for linear regression"
        ndim = x
    else:
        args = (x,)
        if len(x.shape) > 1:
            ndim = x.shape[1]
        else:
            raise ValueError(
                "Consider using kernels.linear for a 1-dimensional data array"
            )
    fixed = {}
    for p_name, p_value in constraints.items():
        if p_name.startswith("beta_"):
            p_name = p_name[len("beta_") :]
        if isinstance(x, pd.DataFrame) and p_name in x.columns:
            index = tuple(x.columns).index(p_name) + 1
        else:
            index = int(p_name)
        fixed[index] = ConstantParameter(p_value)

    if 0 in fixed and not add_intercept:
        raise ValueError(
            "A fixed value is given for the intercept, but `add_intercept` is not True."
        )

    param_indices = range(0 if add_intercept else 1, ndim + 1)
    params = {
        f"beta_{i}": Parameter() if i not in fixed else fixed[i] for i in param_indices
    }

    def _compute(data, **params_from_wrapper):
        sorted_params = [params_from_wrapper[k] for k in params]
        return np.exp((sorted_params * data).sum(axis=1))

    return ParametrizedFunction(
        _compute, *args, **params, fname=exponential_linear_regression.__qualname__
    )


def polynomial_regression(
    x: Union[int, pd.DataFrame, np.ndarray] = 2,
    degree: Union[int, Sequence] = 2,
    **constraints,
) -> ParametrizedFunction:
    """Computes a trend as the sum of the columns in the data to the power of n for n smaller or equal to degree.

    :param x: the number of dimensions or the data the kernel will be computed on. There will be one parameter for each column.
    :param degree: last exponent computed for the given covariates. Can be a list or np array, but if this is the case, the number of
    exponents should be equal to the number of columns of x.
    :param constraints: fixed values for the parameters of the regression. The following constraints are equivalent:
                        'beta_2_2=2', 'beta_cname_2=2', 'cname_2=2'
                        The last two are valid only if data is given as a dataframe with the second column named 'cname'.
    """
    args = ()
    if isinstance(x, int):
        assert x > 0, "Unexpected number of parameters for polynomial regression"
        ndim = x
    else:
        args = (x,)
        if len(x.shape) > 1:
            ndim = x.shape[1]
        else:
            raise ValueError(
                "Consider using kernels.linear for a 1-dimensional data array"
            )
    if isinstance(degree, int):
        assert degree > 0, "This model considers positive power laws only."
        degree = [degree] * ndim
    else:
        assert (
            len(degree) == ndim
        ), "The number of degrees is different than the number of covariates."
    ncols = sum(degree)
    fixed = {}
    for p_name, p_value in constraints.items():
        if p_name.startswith("beta_"):
            p_name = p_name[len("beta_") :]
        match = re.match(r"^(.+)_(\d+)$", p_name)
        if match:
            column, deg = match.groups()
        else:
            raise ValueError(f"Unable to parse parameter constraint: {p_name}")
        if isinstance(x, pd.DataFrame) and column in x.columns:
            column = list(x.columns).index(column) + 1
        fixed[(int(column), int(deg))] = ConstantParameter(p_value)
    params = {}
    for col_idx, max_degree in enumerate(degree):
        for d in range(1, max_degree + 1):
            name = f"beta_{col_idx+1}_{d}"
            params[name] = fixed.get((col_idx + 1, d), Parameter())

    def _compute(data, **params_from_wrapper):
        data = np.array(data)
        data_with_extra_cols = np.zeros(shape=(len(data), ncols))
        extra_col_idx = 0
        for col_idx, max_degree in enumerate(degree):
            for d in range(1, max_degree + 1):
                data_with_extra_cols[:, extra_col_idx] = data[:, col_idx] ** d
                extra_col_idx += 1
        sorted_params = [params_from_wrapper[k] for k in params]
        return (sorted_params * data_with_extra_cols).sum(axis=1)

    return ParametrizedFunction(
        _compute, *args, **params, fname=polynomial_regression.__qualname__
    )


def categories_qualitative(
    x: Collection, fixed_values: dict = None
) -> ParametrizedFunction:
    unique_values = sorted(set(map(str, x)))
    fixed_values = {str(k): v for k, v in (fixed_values or {}).items()}
    parameter = (Parameter() for _ in count())  # generate parameters on demand
    params = {
        value: next(parameter)
        if value not in fixed_values
        else ConstantParameter(fixed_values[value])
        for value in unique_values
    }

    def _compute(data, **params_from_wrapper):
        return type(data)(list(map(lambda v: params_from_wrapper[str(v)], data)))

    return ParametrizedFunction(
        _compute, x, **params, fname=categories_qualitative.__qualname__
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pykelihood', 'pykelihood.visualisation']

package_data = \
{'': ['*']}

install_requires = \
['cachetools>=4.2.1,<5.0.0',
 'numpy>=1.16.5,<2.0.0',
 'pandas>=1.2.1,<2.0.0',
 'scipy>=1.5.3,<2.0.0']

setup_kwargs = {
    'name': 'pykelihood',
    'version': '0.2.0',
    'description': 'Intuitive syntax for complex statistical model fitting and parameter sensitivity analysis.',
    'long_description': 'pykelihood\n==========\n\n\n.. image:: https://badge.fury.io/py/pykelihood.svg\n    :target: https://pypi.org/project/pykelihood/\n\n.. image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n\n-----\nAbout\n-----\n\nPykelihood is a Python package for statistical analysis designed to give more flexibility to likelihood-based inference\nthan is possible with **scipy.stats**. Distributions are designed from an Object Oriented Programming (OOP) point of\nview.\n\nMain features include:\n\n- use any **scipy.stats** distribution, or make your own,\n- fit distributions of arbitrary complexity to your data,\n- add trends of different forms in the parameters of any distribution,\n- condition the likelihood with any form of penalty,\n- profile parameters with a penalised likelihood,\n- fit joint distributions and point processes with self exciting or time dependent intensity,\n- more to come...\n\n\n------------\nInstallation\n------------\n\nUsing pip\n------------------\n\n.. code::\n\n    pip install pykelihood\n\n\nFrom sources\n------------\n\n.. code::\n\n    git clone https://www.github.com/OpheliaMiralles/pykelihood\n\nor\n\n.. code::\n\n    gh repo clone OpheliaMiralles/pykelihood\n\n\n-----\nUsage\n-----\n\nBasics\n------\n\nThe most basic thing you can use ``pykelihood`` for is creating and manipulating distributions as objects.\n\n>>> from pykelihood.distributions import Normal\n>>> n = Normal(1, 2)\n>>> n\nNormal(loc=1.0, scale=2.0)\n\n``n`` is an *object* of type ``Normal``. It has 2 parameters, ``loc`` and ``scale``. They can be accessed like standard\nPython attributes:\n\n>>> n.loc\n1.0\n\nUsing the ``Normal`` object, you can calculate standard values using the same semantics as **scipy.stats**:\n\n>>> n.pdf([0, 1, 2])\narray([0.17603266, 0.19947114, 0.17603266])\n>>> n.cdf([0, 1, 2])\narray([0.30853754, 0.5       , 0.69146246])\n\nOr you can also generate random values according to this distribution:\n\n>>> n.rvs(10)\narray([ 3.31370986,  5.02699468, -0.3573229 ,  1.00460378, -3.26044871,\n        1.86362711, -0.84192901,  0.81132182, -2.03266978,  1.48079944])\n\n\nFitting\n-------\n\nLet\'s generate a larger sample from our previous object:\n\n>>> data = n.rvs(1000)\n>>> data.mean()\n1.025039359276458\n>>> data.std()\n1.9376460645596842\n\nWe can fit a ``Normal`` distribution to this data, which will return another ``Normal`` object:\n\n>>> n.fit(data)\nNormal(loc=1.0250822420920338, scale=1.9376400770300832)\n\nAs you can see, the values are slightly different from the moments in the data.\nThis is due to the fact that the ``fit`` method returns the Maximum Likelihood Estimator (MLE)\nfor the data, and is thus the result of an optimisation (using **scipy.optimize**).\n\nWe can also fix the value for some parameters if we know them:\n\n>>> n.fit(data, loc=1)\nNormal(loc=1.0, scale=1.9377929687500024)\n\nTrend fitting\n*************\n\nOne of the most powerful features of ``pykelihood`` is the ability to fit arbitrary distributions.\nFor instance, say our data has a linear trend with a very little gaussian noise we would like to capture:\n\n>>> import numpy as np\n>>> data = np.linspace(-1, 1, 365) + np.random.normal(0, 0.001, 365)\n>>> data[:10]\narray([-0.99802364, -0.99503679, -0.98900434, -0.98277981, -0.979487  ,\n       -0.97393519, -0.96853445, -0.96149152, -0.95564004, -0.95054887])\n\nIf we try to fit this without a trend, the resulting distribution will miss out on most of the information:\n\n>>> n.fit(data)\nNormal(loc=-3.6462053656578005e-05, scale=0.5789668679237372)\n\nLet\'s fit a ``Normal`` distribution with a trend in the loc parameter:\n\n>>> from pykelihood import kernels\n>>> n.fit(data, loc=kernels.linear(np.arange(365)))\nNormal(loc=linear(a=-1.0000458359290572, b=0.005494714384381866), scale=0.0010055323717468906)\n\n``kernels.linear(X)`` builds a linear model in the form *a + bX* where *a* and *b* are parameters to\nbe optimised for, and *X* is some covariate used to fit the data. If we assume the data were daily observations,\nthen we find all the values we expected: *-1* was the value on the first day, *0.05* was the daily increment\n(*2 / 365 = 0.05*), and there was a noise with std deviation *0.001*.\n\n\nParameter profiling\n*******************\n\nLikelihood based inference relies on parameter estimation. This is why it\'s important to quantify the sensitivity of a\nchosen model to each of those parameters. The ``stats_utils`` module in ``pykelihood`` includes the ``Likelihood``\nclass that allows to link a model to a set of observations by providing goodness of fit metrics and profiles for all\nparameters.\n\n>>> from pykelihood.stats_utils import Likelihood\n>>> from pykelihood.distributions import GEV\n>>> fitted_gev = GEV.fit(data, loc=kernels.linear(np.linspace(-1, 1, len(data))))\n>>> ll = Likelihood(fitted_gev, data, inference_confidence=0.99) # level of confidence for tests\n>>> ll.AIC  # the standard fit is without trend\n{\'AIC MLE\': -359.73533182968777, \'AIC Standard MLE Fit\': 623.9896838880583}\n>>> ll.profiles.keys()\ndict_keys([\'loc_a\', \'loc_b\', \'scale\', \'shape\'])\n>>> ll.profiles["shape"].head(5)\n      loc_a     loc_b     scale     shape   likelihood\n0 -0.000122  1.000812  0.002495 -0.866884  1815.022132\n1 -0.000196  1.000795  0.001964 -0.662803  1882.043541\n2 -0.000283  1.000477  0.001469 -0.458721  1954.283256\n3 -0.000439  1.000012  0.000987 -0.254640  2009.740282\n4 -0.000555  1.000016  0.000948 -0.050558  1992.812843\n\nConfidence intervals can be computed for specified metrics:\n\n>>> def metric(gev): return gev.loc()\n>>> ll.confidence_interval(metric)\n[-4.160287666875364, 4.7039931595123825]\n\n\n------------\nContributing\n------------\n\n`Poetry <http://python-poetry.org>`_ is used to manage ``pykelihood``\'s dependencies and build system. To install\nPoetry, you can refer to the `installation instructions <https://python-poetry.org/docs/#installation>`_, but it boils\ndown to running:\n\n.. code::\n\n    curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python\n\n\n\nTo configure your environment to work on pykelihood, run:\n\n.. code-block::\n\n    git clone https://www.github.com/OpheliaMiralles/pykelihood  # or any other clone method\n    cd pykelihood\n    poetry install\n\nThis will create a virtual environment for the project and install the required dependencies. To activate the virtual\nenvironment, be sure to run :code:`poetry shell` prior to executing any code.\n\nWe also use the `pre-commit <https://pre-commit.com>`_ library which adds git hooks to the repository. These must be installed with:\n\n.. code::\n\n   pre-commit install\n\nSome parts of the code base use the `matplotlib <https://matplotlib.org/>`_ and\n`hawkeslib <https://hawkeslib.readthedocs.io/en/latest/index.html>`_ package, but are for now not required to run most\nof the code, including the tests.\n\nTests\n-----\n\nTests are run using `pytest <https://docs.pytest.org/en/stable/>`_. To run all tests, navigate to the root folder or the\n``tests`` folder and type :code:`pytest`.\n',
    'author': 'Ophélia Miralles',
    'author_email': 'ophelia.miralles@epfl.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/OpheliaMiralles/pykelihood',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from collections import ChainMap
from functools import partial
from typing import Any, Callable, Dict, Iterable, Tuple, TypeVar, Union

from pykelihood.utils import flatten_dict

_T = TypeVar("_T")


class Parametrized(object):
    params_names: Tuple[str]

    def __init__(self, *params: Union[Parametrized, Any]):
        self._params = tuple(
            Parameter(p) if not isinstance(p, Parametrized) else p for p in params
        )

    def _build_instance(self, **new_params):
        sorted_params = [new_params[p_name] for p_name in self.params_names]
        return type(self)(*sorted_params)

    @property
    def params(self) -> Tuple[Parametrized]:
        return self._params

    @property
    def param_dict(self) -> Dict[str, Parametrized]:
        return dict(zip(self.params_names, self.params))

    @property
    def flattened_params(self) -> Tuple[Parametrized]:
        return tuple(p_ for p in self.params for p_ in p.params)

    @property
    def flattened_param_dict(self) -> Dict[str, Parametrized]:
        p_dict = flatten_dict(self._flattened_param_dict_helper())
        return {"_".join(names): value for names, value in p_dict.items()}

    def _flattened_param_dict_helper(self):
        return {
            name: value._optimisation_param_dict_helper()
            for name, value in self.param_dict.items()
        }

    @property
    def optimisation_params(self) -> Tuple[Parametrized]:
        return tuple(p_ for p in self.params for p_ in p.optimisation_params)

    @property
    def optimisation_param_dict(self) -> Dict[str, Parametrized]:
        p_dict = flatten_dict(self._optimisation_param_dict_helper())
        return {"_".join(names): value for names, value in p_dict.items()}

    def _optimisation_param_dict_helper(self):
        return {
            name: value._optimisation_param_dict_helper()
            for name, value in self.param_dict.items()
            if not isinstance(value, ConstantParameter)
        }

    def __repr__(self):
        args = [f"{a}={v!r}" for a, v in zip(self.params_names, self._params)]
        return f"{type(self).__name__}({', '.join(args)})"

    def with_params(self, params: Iterable = None, **named_params):
        if params is not None and named_params:
            raise ValueError("Please only use one way to provide values to parameters.")
        if params is not None:
            params = iter(params)
            new_params = {}
            for p_name, param in self.param_dict.items():
                new_params[p_name] = param.with_params(params)
        else:
            new_params = {}
            for p_name, param in self.param_dict.items():
                this_param_values = {}
                for name, value in named_params.items():
                    # check if this value is for this parameter
                    if name == p_name:
                        new_params[name] = value
                        # no need to look further
                        break
                    elif name.startswith(p_name + "_"):
                        remaining = name[len(p_name + "_") :]
                        this_param_values[remaining] = value
                else:  # nobreak
                    if this_param_values:
                        new_params[p_name] = param.with_params(**this_param_values)
            # use current values if none were given
            new_params = ChainMap(new_params, self.param_dict)
        return self._build_instance(**new_params)

    def __getattr__(self, param: str) -> Parametrized:
        try:
            idx = self.params_names.index(param)
        except ValueError:
            raise AttributeError(f"No parameter {param} in {type(self).__name__}")
        return self._params[idx]


class Parameter(float, Parametrized):
    def __new__(cls, x=0.0):
        return float.__new__(cls, x)

    def __init__(self, *args, **kwargs):
        pass

    @property
    def params(self):
        return (self,)

    @property
    def optimisation_params(self):
        return (self,)

    def _optimisation_param_dict_helper(self, prefixes=()):
        return self

    def with_params(self, params):
        param = next(iter(params))
        if isinstance(param, ConstantParameter):
            return param
        return type(self)(param)

    def __call__(self):
        return self

    def __repr__(self):
        return float.__repr__(self)

    def __getattr__(self, item):
        raise AttributeError


class ConstantParameter(Parameter):
    def with_params(self, params):
        return self

    @property
    def optimisation_params(self):
        # do not show up in the params
        return ()


class ParametrizedFunction(Parametrized):
    def __init__(self, f: Callable, *args, fname=None, **params: Parametrized):
        self._init_args = args
        self.f = partial(f, *args)
        self.original_f = f
        super(ParametrizedFunction, self).__init__(*params.values())
        self.params_names = tuple(params.keys())
        self.fname = fname or f.__qualname__
        self._mul = 1

    def __call__(self, *args, **kwargs):
        if not args and not kwargs:
            return self._mul * self.f(**self.param_dict)
        return self._mul * self.original_f(*args, **kwargs, **self.param_dict)

    def __getattr__(self, param):
        try:
            return super(ParametrizedFunction, self).__getattr__(param)
        except AttributeError:
            raise AttributeError(f"No parameter {param} in {self.f.func.__qualname__}")

    def __repr__(self):
        args = [f"{a}={v!r}" for a, v in zip(self.params_names, self._params)]
        return f"{self.fname}({', '.join(args)})"

    def _build_instance(self, **new_params):
        return type(self)(self.original_f, *self._init_args, **new_params)



from threading import RLock

_NOT_FOUND = object()


class cached_property:
    def __init__(self, func):
        self.func = func
        self.attrname = None
        self.__doc__ = func.__doc__
        self.lock = RLock()

    def __set_name__(self, owner, name):
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                f"({self.attrname!r} and {name!r})."
            )

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it."
            )
        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(self.attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            with self.lock:
                # check if another thread filled cache while we awaited lock
                val = cache.get(self.attrname, _NOT_FOUND)
                if val is _NOT_FOUND:
                    val = self.func(instance)
                    try:
                        cache[self.attrname] = val
                    except TypeError:
                        msg = (
                            f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                            f"does not support item assignment for caching {self.attrname!r} property."
                        )
                        raise TypeError(msg) from None
        return val
